/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include "nSX2.h"
#include "PS2Edefs.h"

#include "DmaC.h"

/* must create function to avoid this kind of declaration */
//#include "../IntrC/IntrC.h"

_u64 nb_dma_transfert = 0;

/*
Transfer  n ( Ps2Bios.DmaC[ DMA_CH2 ].size  ) QuadWord from
Address ( Ps2Bios.DmaC[ DMA_CH2 ].memAdd ) to the GS plugin
*/
//(Info)	0022e954:[DMAC][GIF][NORMAL] Add:0x01fffde0 S:20480 


void DMAC_GIF_Transfer_Normal_Mode( void ) {
	_u32 	*DmaAdd;

	/* Check if Addr is inside the SPR Memory */
	if ( ((_u32)DmaC.Channel[ DMA_CH2 ]->memAdd > (_u32)0x81FFFFFF ) ||
	     ( ((_u32)DmaC.Channel[ DMA_CH2 ]->memAdd & (_u32)0xFFFF0000 ) == 0x80000000 )
		) {
		/*Find the Adress Pointer and pass it to the GS Pluggin*/
		DmaAdd = (  _u32 * )	&SPRMemory [ DmaC.Channel[ DMA_CH2 ]->memAdd & 0x3FFF];	

	}
	else {
	     _u32 start = DmaC.Channel[ DMA_CH2 ]->memAdd & 0x01FFFFFF ;
		/*Find the Adress Pointer and pass it to the GS Pluggin*/
		if ( (start + DmaC.Channel[ DMA_CH2 ]->size) > ((_u32)0x01FFFFFF )) {
             DmaC.Channel[ DMA_CH2 ]->size = (_u32)0x01FFFFFF - start;
         }    

		DmaAdd = (  _u32 * )	&MipsMemory [ start ];	
		
	}
	
	GIF_PATH3_Transfert( DmaAdd, (_u32 ) ( DmaC.Channel[ DMA_CH2 ]->size / 4 )) ;	
}



void DMAC_GIF_Transfer ( void ) {

	char 	DmaTempText[256];
 
	/* if the control value is START */
	if ( ( DmaC.Channel[ DMA_CH2 ]->ctrl & 0x100 ) == 0x100 ) {
	
		/*Normal Transfert or Chained Transfer*/
		switch ( DmaC.Channel[ DMA_CH2 ]->ctrl & 0xFF ) {
		
		/*Normal Transfert*/
			/* 	From Memory ( Not effective here ), Mode Normal, 
				No Adress pushed by call Tag , 
				Transfers DMATAG ( not effective in Normal Mode )
				Disable IRQ bit of DMATAG
			*/
			case 0x41 :	

			/* 	From Memory ( Not effective here ), Mode Normal, 
				No Adress pushed by call Tag , 
				No Transfers DMATAG 
				Disable IRQ bit of DMATAG
			*/
			case 0x01 : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][GIF][NORMAL] Add:0x%8.8lx S:%d", (long)MipsCpu.PC,
						(long)DmaC.Channel[ DMA_CH2 ]->memAdd, ( long ) DmaC.Channel[ DMA_CH2 ]->size );
#endif

				/* This alignement is needed */
				DmaC.Channel[ DMA_CH2 ]->size = DmaC.Channel[ DMA_CH2 ]->size * 4;
				DMAC_GIF_Transfer_Normal_Mode();
				INTC_RaiseIntByDMA ( DMA_CH2 );
				
			break;
			}
 			
		/*Chained Transfert*/
			/* 	From Memory ( Not effective here ), Mode Chain, 
				No Adress pushed by call Tag , 
				Transfers DMATAG 
				Disable IRQ bit of DMATAG
			*/
			case 0x45 : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][GIF][CHAIN][WITH_TAG] Add:0x%8.8lx ", (long)MipsCpu.PC,
						(long)DmaC.Channel[ DMA_CH2 ]->tag  );
#endif
				/* Didn't work good if I transfer the Tag 
					GS plugin size is 128bit unit , maybe it's the problem

					or maybe I must remove 1 QWORD from the size ?
					Have to test !
				*/
				DMAC_Transfer_SourceChain_Mode( DMA_CH2 , DISABLE_TRANSFERT_TAG ,DISABLE_TAG_IRQ, DMAC_GIF_Transfer_Normal_Mode );

			break;
			}


			/* 	From Memory ( Not effective here ), Mode Chain, 
				No Adress pushed by call Tag , 
				No Transfers DMATAG 
				Disable IRQ bit of DMATAG
			*/
			case 0x05 : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][GIF][CHAIN][NO_TAG] Add:0x%8.8lx ", (long)MipsCpu.PC,
						(long)DmaC.Channel[ DMA_CH2 ]->tag  );
#endif
				DMAC_Transfer_SourceChain_Mode( DMA_CH2 , DISABLE_TRANSFERT_TAG ,DISABLE_TAG_IRQ, DMAC_GIF_Transfer_Normal_Mode );
			
			break;
			}

			/* 	To Memory ( Not effective here ), Mode Chain, 
				No Adress pushed by call Tag , 
				No Transfers DMATAG 
				Enable IRQ bit of DMATAG
			*/
			case 0x84 : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][GIF][CHAIN][NO_TAG] Add:0x%8.8lx ", (long)MipsCpu.PC,
						(long)DmaC.Channel[ DMA_CH2 ]->tag  );
#endif
				DMAC_Transfer_SourceChain_Mode( DMA_CH2 , DISABLE_TRANSFERT_TAG ,ENABLE_TAG_IRQ, DMAC_GIF_Transfer_Normal_Mode );
			
			break;
			}


			
			
			/*Other Transfer Mode GS Not Implemented*/
			default: {
#ifdef EMU_LOG
				EMU_LOG("%8.8lx:[DMAC][GIF][MODE_?] mode:%2.2llx ", (long)MipsCpu.PC,(DmaC.Channel[ DMA_CH2 ]->ctrl & 0xFF ) );
#endif
			}
			break;

		
											
		}
	
		if ( debug_dma == 1 ) PrintToDmaOutput (  " \n" );


		/*This call write the end of the dma transfert
		with writing 0 to the bit start of the DM2 channel Control */
		DmaC.Channel[ DMA_CH2 ]->ctrl = DmaC.Channel[ DMA_CH2 ]->ctrl &  0xFFFFFEFF ;
		DmaC.Channel[ DMA_CH2 ]->size = (_u32)0;
		/*  */
		INTC_RaiseIntByCause ( INTC_GS );

	
	}
				
}









