/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include "nSX2.h"
#include "PS2Edefs.h"

#include "DmaC.h"


 


/*
NOTE :
All size is 32 bits word count ( not Qwords )

IMPORTANT : CHECK WHAT I NEED TO DO WITH : InterruptRequest
*/

void DMAC_Transfer_SourceChain_Mode( _u8 Channel , _u8 TransfertTag , _u8 EnableIRQ ,void (*fNormalMode)(void) ) {

	_u32 	*DmaAdd;
	_u32	tempAddr;
	char 	DmaTempText[256];

	_u128 	temp;

	_u16    Tag_ID_FLAG;
	_u8		InterruptRequest;
    _u8     DChainTransfertCompleted;

	
	_u32	AdressAfterDATA;
	_u8		PacketCount = 1; 

	/*Chained Mode Initialisation*/
	DChainTransfertCompleted = FALSE;


	/*Transfert until end */
	while ( DChainTransfertCompleted != TRUE ) {
		
		/*Read Tag from the Tag Adress*/
	    MipsMemoryRead ( &temp  , DATA_DW , DmaC.Channel[ Channel ]->tag , 0 ) ;

#ifdef DMAC_LOG
		DMAC_LOG("%8.8lx:[DMAC][CHAIN_TAG] Add:0x%8.8lx  Val:0x%8.8lx", (long)MipsCpu.PC, DmaC.Channel[ Channel ]->tag ,temp.lo);
#endif

		/* Update Channel Control Register ( TAG part )*/
		DmaC.Channel[ Channel ]->ctrl = ( DmaC.Channel[ Channel ]->ctrl & 0xFFFF ) | ( temp.lo & 0xFFFF0000 );
	
		/*Decode Tag*/
	    DmaC.Channel[ Channel ]->size		= ( temp.lo   & 0xFFFF); 	/* In QWORDS unit */     						
	    Tag_ID_FLAG							= ( temp.lo >> 28 ) & 0x7 ; 
		InterruptRequest					= ( temp.lo >> 31 ) & 0x1 ;  
	    DmaC.Channel[ Channel ]->memAdd		= (_u32)(( temp.lo >> 32 )& 0xFFFFFFFF); 

		/* All Size is 32bit unit , NOW */
		DmaC.Channel[ Channel ]->size = DmaC.Channel[ Channel ]->size * 4;

#ifdef DMAC_LOG
		DMAC_LOG("%8.8lx:[DMAC][CHAIN_TAG] ID:%d, S:%d, Add:0x%8.8lx", (long)MipsCpu.PC, Tag_ID_FLAG, 
				DmaC.Channel[ Channel ]->size, (long)DmaC.Channel[ Channel ]->memAdd);
#endif

		/* if the TAG IRQ is enable and if the Transfert of the Tag is done */
		if ( PacketCount == 0 ) Tag_ID_FLAG = 0xFF;
		
	 	/*Execute Tag*/
	    switch ( Tag_ID_FLAG ) {

			/* Emulator Internal TAG for Interrupt */
			case 0xFF: {
				
				/* Interrupt : Tag interrupt */
				if ( EnableIRQ == ENABLE_TAG_IRQ ) 			
					INTC_RaiseIntByDMA ( Channel );  


				DChainTransfertCompleted = TRUE;
			break;
			}


	    	/*
	    	Transfers the QWC qword from the ADDR field, clears the
			Dn_CHCR.STR to 0, and ends transfer.
	    	*/
	        case DMA_SRC_CHAIN_MODE_REFE : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][REFE]", (long)MipsCpu.PC );
#endif
				fNormalMode();	
				
                /*End Chained Transfer*/                                 	        	
	        	PacketCount--;
	        break;
	        }


			/*
			Transfers the QWC qword following the tag and reads the
			succeeding qword as the next tag.
			*/
	        case DMA_SRC_CHAIN_MODE_CONT : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][CNT]", (long)MipsCpu.PC );
#endif

				AdressAfterDATA = DmaC.Channel[ Channel ]->tag  + 16 + (  ( DmaC.Channel[ Channel ]->size  )  * 4 ) ;
	

				if ( TransfertTag == TRUE ) {
	        		/*Compute Address after the Tag*/
	        		DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 8 ;
					/* Adjust the size ( Transfert 128 <--> 64 bits of the tag )*/
					DmaC.Channel[ Channel ]->size = DmaC.Channel[ Channel ]->size + 2 ;
				        	
				} else {

					DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 16 ;
				}
			
				fNormalMode();


   	        	/*Compute the Adress after the DATA*/
				DmaC.Channel[ Channel ]->tag = AdressAfterDATA;

	        	
	        break;
	        }
	        

			/*
			Transfers the QWC qword following the tag and 
			reads the  qword of the ADDR field as the next tag
			*/
	        case DMA_SRC_CHAIN_MODE_NEXT : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][NEXT]", (long)MipsCpu.PC );
#endif

	        	/*Save the ADDR Field*/
	        	tempAddr = DmaC.Channel[ Channel ]->memAdd ;
	
				if ( TransfertTag == TRUE ) {
	        		/*Compute Address after the Tag*/
	        		DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 8 ;
					/* Adjust the size ( Transfert 128 <--> 64 bits of the tag )*/
					DmaC.Channel[ Channel ]->size = DmaC.Channel[ Channel ]->size + 2 ;
				        	
				} else {

					DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 16 ;
				}

	
				fNormalMode();	
    			
				DmaC.Channel[ Channel ]->tag = tempAddr ;	 
      	
	        break;
	        }


			/*
			Transfers the QWC qword from the ADDR field and reads the
			qword following the tag as the next tag.
			*/       
	        case DMA_SRC_CHAIN_MODE_REF : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][REF]", (long)MipsCpu.PC );
#endif
			
				fNormalMode();

	        	/*Compute the Adress after the TAG*/
	        	DmaC.Channel[ Channel ]->tag = DmaC.Channel[ Channel ]->tag  +  16 ;

    
	        break;
	        }


			/*
			Transfers the QWC qword from the ADDR field while
			controlling stalls and reads the qword following the tag as the
			next tag.
			Effective only on the VIF1, GIF, and SIF1 channels.	        
			*/
	        case DMA_SRC_CHAIN_MODE_REFS : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][REFS]", (long)MipsCpu.PC );
#endif
				fNormalMode();	

	        	/*Compute the Adress after the TAG*/
	        	DmaC.Channel[ Channel ]->tag = DmaC.Channel[ Channel ]->tag  +  16 ;
	        	

	        break;
	        }



			/*
			Transfers the QWC qword following the tag, pushes the next
			field into the Dn_ASR register, and reads the qword of the
			ADDR field as the next tag.
			Effective only on the VIF0, VIF1, and GIF channels.
			Addresses can be pushed up to 2 levels.
			*/	        
	        case DMA_SRC_CHAIN_MODE_CALL : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][CALL]", (long)MipsCpu.PC );
#endif
 	        	
				/*Save the ADDR Field*/
	        	tempAddr = DmaC.Channel[ Channel ]->memAdd ;


				/* Read the Next field as a Tag and store it */
				AdressAfterDATA = (_u32)DmaC.Channel[ Channel ]->tag + 16  + (_u32)(DmaC.Channel[ Channel ]->size * 4) ;

				
				if ( DmaC.Channel[ Channel ]->asr0 == 0 ) {
					DmaC.Channel[ Channel ]->asr0 = AdressAfterDATA;
				} else {
					DmaC.Channel[ Channel ]->asr1 = AdressAfterDATA;	
				}
				


				if ( TransfertTag == TRUE ) {
	        		/*Compute Address after the Tag*/
	        		DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 8 ;
					/* Adjust the size - Only needed when TAG is send*/
					DmaC.Channel[ Channel ]->size = DmaC.Channel[ Channel ]->size + 2 ;
				        	
				} else {

					DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 16 ;
				}
				
				fNormalMode();


				DmaC.Channel[ Channel ]->tag = 	tempAddr;


	        
	        break;
	        }



			/*	
			Transfers the QWC qword following the tag and reads the
			qword of the field popped from the Dn_ASR register as the
			next tag.
			Transfers the QWC qword following the tag, clears the
			Dn_CHCR.STR to 0, and ends transfer when there is no
			pushed address.
			Effective only on the VIF0, VIF1, and GIF channels.	        
			*/
	        case DMA_SRC_CHAIN_MODE_RET : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][RET]", (long)MipsCpu.PC );
#endif


				if ( TransfertTag == TRUE ) {
	        		/*Compute Address after the Tag*/
	        		DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 8 ;
					/* Adjust the size ( Transfert 128 <--> 64 bits of the tag )*/
					DmaC.Channel[ Channel ]->size = DmaC.Channel[ Channel ]->size + 2 ;
				        	
				} else {

					DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 16 ;
				}

				fNormalMode();


				if ( DmaC.Channel[ Channel ]->asr1 != 0 ) {
					DmaC.Channel[ Channel ]->tag = DmaC.Channel[ Channel ]->asr1;
					DmaC.Channel[ Channel ]->asr1 = 0;
			

				} else {
					if ( DmaC.Channel[ Channel ]->asr0 != 0 ) {
						DmaC.Channel[ Channel ]->tag = DmaC.Channel[ Channel ]->asr0;
						DmaC.Channel[ Channel ]->asr0 = 0;
					} else {

  	        			/*End Chained Transfer*/ 
	        			PacketCount--;
					}
				}

      	

	        break;
	        }



			/*
			Transfers the QWC qword following the tag, clears the
			Dn_CHCR.STR to 0, and ends the transfer.	        
			*/
	        case DMA_SRC_CHAIN_MODE_END : {
#ifdef DMAC_LOG
				DMAC_LOG("%8.8lx:[DMAC][CHAIN][END]", (long)MipsCpu.PC );
#endif
	        	        	
				if ( TransfertTag == TRUE ) {
	        		/*Compute Address after the Tag*/
	        		DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 8 ;
					/* Adjust the size ( Transfert 128 <--> 64 bits of the tag )*/
					DmaC.Channel[ Channel ]->size = DmaC.Channel[ Channel ]->size + 2 ;
				        	
				} else {

					DmaC.Channel[ Channel ]->memAdd = DmaC.Channel[ Channel ]->tag + 16 ;
				}
	        	
				fNormalMode();
	
  	        	/*End Chained Transfer*/ 
	        	PacketCount--;
	        break;
	        }


			/*
	        Undefined TAG
			*/
	        default : {
#ifdef EMU_LOG
				EMU_LOG("%8.8lx:[DMAC][CHAIN][TAG_?] n %d", (long)MipsCpu.PC,Tag_ID_FLAG );
#endif
	        	DChainTransfertCompleted = TRUE;
	        break;
	        }
	        
	        
	    }//End Switch DMA Chained TAG
	
	}//End While Dma Chained Transfer
}



