/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>


#include "nSX2.h"
#include "PS2Edefs.h"

#include "DmaC.h"


DMAController 		DmaC;

 
void DMAC_Log ( char * ftext  , ...   ) {
	char DMAC_LogText[ FILE_NAME_LENGHT ];

	if ( debug_dma == 1 ) {

	    va_list ap;
   		va_start (ap, ftext);
    	vsprintf (DMAC_LogText, ftext, ap);
    	va_end (ap);

		sprintf(DMAC_LogText,"%s\n",DMAC_LogText);
		PrintToDmaOutput (  DMAC_LogText ); 

	}
}



/* DMA Controller Initialisation */
void	DMAC_Init( void ) {

#ifdef EMU_LOG
	EMU_LOG("[DMAC]Initialisation");
#endif

	/* Emulator Initialisation */
	DmaC.Reg = ( DmaControllerRegisters * ) &SpecialAreaMemory[   ( DMAC_CTRL & 0xFFFF )  ];

	DmaC.Channel[ DMA_CH0 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D0_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH1 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D1_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH2 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D2_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH3 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D3_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH4 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D4_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH5 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D5_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH6 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D6_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH7 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D7_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH8 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D8_CHCR & 0xFFFF ) ];
	DmaC.Channel[ DMA_CH9 ] = ( DmaChannelRegisters * ) &SpecialAreaMemory[ ( D9_CHCR & 0xFFFF ) ];

	/* DMA Controller Initialisation */
	DmaC.Reg->stat = 0x00000000;	

	DmaC.Reg->ctrl = 0x00000001;
	DmaC.EnableAllDMAs = 1;

}




/* Write to DMA Controller/Channel Registers */
void	DMAC_Write ( _u64 fPhysicalAddress , _u128 fVal , _u8 fDataType ) {

	_u32	boucle;
	_u64	RealAdd;

	/*
		if emu write to the DMAC STATUS , need to perform bit operation.
		For all other value, simply write
	*/
	if ( fPhysicalAddress != DMAC_STAT ) {

		RealAdd 		= fPhysicalAddress & 0xFFFF;

		switch ( fDataType ) {

			case DATA_QUAD:	MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.lo, RealAdd );
							MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.hi, RealAdd + 8 );break;

				
			case DATA_DW:	MipsMemoryWriteDW	( SpecialAreaMemory, fVal.lo, RealAdd );break;
				
			case DATA_W:	MipsMemoryWriteW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

			case DATA_HW:	MipsMemoryWriteHW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

			case DATA_B:	MipsMemoryWriteB	( SpecialAreaMemory, fVal.lo, RealAdd );break;
										
			default:		EMU_LOG("%8.8lx:[DMAC][WR][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
							MipsCpu.halt = 1;break;
		}


	}


	/* 
		if DMACLOG isn't defined, only the D?_CHCR ( DMA CHANNEL  ? CONTROL ) are checked and
			the DMAC_STAT , DMAC_CTRL
		else same as above with Log of all others registers
	*/
	switch ( fPhysicalAddress  ) {

/* DMA CH0 REGISTERS (Linked to VIF0) */
		case D0_CHCR:	//Ch0 Channel Control
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			if ( DmaC.EnableAllDMAs ) DMAC_VIF0_Transfer ();			
		break;

#ifdef DMAC_LOG
 		case D0_MADR:  	//Ch0 Memory Address	
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D0_QWC:	//Ch0 Quad word count
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_TADG : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
				
		case D0_TADG:	//Ch0 Tag address 
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_TADG : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
						
		case D0_ASR0:	//CH0 Address stack 0
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
			
		case D0_ASR1:	//Ch0 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D0_SADR:	//Ch0 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][WR]D0_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif

/* DMA CH1 REGISTERS (Linked to VIF1) */
		case D1_CHCR: 	//Ch1 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

			if ( DmaC.EnableAllDMAs ) DMAC_VIF1_Transfer ();
		break;


#ifdef DMAC_LOG		
		case D1_MADR: 	//Ch1 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D1_QWC:	//Ch1 Quad word count		
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D1_TADR:	//Ch1 Tag address			
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D1_ASR0:	//CH1 Address stack 0			
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D1_ASR1:	//CH1 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D1_SADR:	//CH1 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][WR]D1_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH2 REGISTERS (Linked to GIF) */
		case D2_CHCR: 	//Ch2 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			
			if ( DmaC.EnableAllDMAs ) DMAC_GIF_Transfer ();

		break;


#ifdef DMAC_LOG		
		case D2_MADR: 	//Ch2 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D2_QWC:	//Ch2 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D2_TADR:	//Ch2 Tag address			
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D2_ASR0:	//CH2 Address stack 0			
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D2_ASR1:	//CH2 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D2_SADR:	//CH2 SPR Address			
			DMAC_LOG("%8.8lx:[DMAC][WR]D2_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH3 REGISTERS (Linked to IPU (FROM IPU)) */
		case D3_CHCR: 	//Ch3 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			
			if ( DmaC.EnableAllDMAs ) DMAC_FromIPU_Transfer();
		break;


#ifdef DMAC_LOG		
		case D3_MADR: 	//Ch3 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D3_QWC:	//Ch3 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D3_TADR:	//Ch3 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D3_ASR0:	//Ch3 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D3_ASR1:	//Ch3 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D3_SADR:	//Ch3 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D3_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH4 REGISTERS (Linked to IPU (TO IPU)) */
		case D4_CHCR: 	//Ch4 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			if ( DmaC.EnableAllDMAs ) DMAC_ToIPU_Transfer();
		break;

#ifdef DMAC_LOG		
		case D4_MADR: 	//Ch4 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D4_QWC:	//Ch4 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D4_TADR: 	//Ch4 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D4_ASR0: 	//Ch4 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D4_ASR1: 	//Ch4 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D4_SADR: 	//Ch4 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D4_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH5 REGISTERS (Linked to SIF0) */
		case D5_CHCR: 	//Ch5 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D5_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			
			if ( DmaC.EnableAllDMAs ) DMAC_SIF0_Transfer ( );
		break;


#ifdef DMAC_LOG		
		case D5_MADR: 	//Ch5 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D5_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D5_QWC:	//Ch5 Quad word count		
			DMAC_LOG("%8.8lx:[DMAC][WR]D5_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH6 REGISTERS (Linked to SIF1) */
		
		case D6_CHCR: 	//Ch6 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D6_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			if ( DmaC.EnableAllDMAs ) DMAC_SIF1_Transfer();
		break;


#ifdef DMAC_LOG		
		case D6_MADR: 	//Ch6 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D6_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D6_QWC:	//Ch6 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D6_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
	
		case D6_TADR: 	//Ch6 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D6_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH7 REGISTERS (Linked to SIF2) */
		case D7_CHCR: 	//Ch7 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D7_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

			if ( DmaC.EnableAllDMAs ) DMAC_SIF2_Transfer();	

		break;

#ifdef DMAC_LOG		
		case D7_MADR: 	//Ch7 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D7_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D7_QWC:	//Ch7 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D7_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH8 REGISTERS (Linked to SPR  (from SCRATCH PAD) */
		case D8_CHCR: 	//Ch8 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
	
			if ( DmaC.EnableAllDMAs ) DMAC_FromSPR_Transfer ();
		break;

#ifdef DMAC_LOG		
		case D8_MADR: 	//Ch8 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D8_QWC:	//Ch8 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D8_TADR:	//Ch8 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D8_ASR0:	//Ch8 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D8_ASR1:	//Ch8 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D8_SADR:	//Ch8 SPR address			
			DMAC_LOG("%8.8lx:[DMAC][WR]D8_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


/* DMA CH9 REGISTERS (Linked to SPR  (To SCRATCH PAD) */


		case 0x1000D401:
		case D9_CHCR: 	//Ch9 Channel Control		
#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			if ( DmaC.EnableAllDMAs ) DMAC_ToSPR_Transfer();
		break;


#ifdef DMAC_LOG		
		case D9_MADR: 	//Ch9 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_MADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case D9_QWC:	//Ch9 quad word count			
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_QWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
	
		case D9_TADR: 	//Ch9 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_TADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D9_ASR0: 	//Ch9 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case D9_ASR1: 	//Ch9 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
	
		case D9_SADR:		//Ch9 SPR address			
			DMAC_LOG("%8.8lx:[DMAC][WR]D9_SADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif

 

/* DMAC REGISTERS */
	
		case DMAC_CTRL:   //DMAC Control           
#ifdef DMAC_LOG	
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_CTRL : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->ctrl);
#endif

			DmaC.EnableAllDMAs = ( DmaC.Reg->ctrl & 0x1 );
		
		
		break;

		
		case DMAC_STAT:   //DMAC Status           

#ifdef DMAC_LOG
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_STAT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

			/* MAYBE DO THE SAME THAN INTC HERE */
			//Cleared When 1 is written
			DmaC.Reg->stat = DmaC.Reg->stat & ~(fVal.lo & 0xFFFF);
		
			

			//Reversed When 1 is written
			for ( boucle = 16 ;boucle < 32 ; boucle++ ) {
				if ( (fVal.lo & ( 1 << boucle )) == ( 1 << boucle ) ) {

					if ( ( DmaC.Reg->stat & ( 1 << boucle ) ) == ( 1 << boucle ) ) {
						DmaC.Reg->stat = DmaC.Reg->stat & ~( 1 << boucle );
					}else {
						DmaC.Reg->stat = DmaC.Reg->stat | ( 1 << boucle );
					}
				}
			}

		break;

#ifdef DMAC_LOG		
		case DMAC_PCR:    //DMAC Priority control           
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_PCR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case DMAC_SQWC:   //DMAC skip quad word           
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_SQWC : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case DMAC_RBSR:    //DMAC ring buffer size          
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_RBSR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case DMAC_RBOR:    //DMAC Ring buffer offset          
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_RBOR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		
		case DMAC_STADR:  //DMA stall address           
			DMAC_LOG("%8.8lx:[DMAC][WR]DMAC_STADR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif

#ifdef DMAC_LOG			
		default: {
#ifdef EMU_LOG
			EMU_LOG("%8.8lx:[DMAC][WR]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
#endif
		break;
		}
#endif
		
	}
}







/* Read from DMA Controller/Channel Registers */
void	DMAC_Read ( _u64 fPhysicalAddress , _u128 * fVal , _u8 fDataType ) {

	_u64 	RealAdd;

#ifdef DMAC_LOG	
	switch ( fPhysicalAddress  ) {
/* DMA CH0 REGISTERS (Linked to VIF0) */
		case D0_CHCR:	//Ch0 Channel Control
			DMAC_LOG("%8.8lx:[DMAC][RD]D0_CHCR : 0x%8.8lx ", (long)MipsCpu.PC, DmaC.Channel[ DMA_CH0 ]->ctrl);
		break;
 		
		case D0_MADR:  	//Ch0 Memory Address	
			DMAC_LOG("%8.8lx:[DMAC][RD]D0_MADR : 0x%8.8lx ", (long)MipsCpu.PC, DmaC.Channel[ DMA_CH0 ]->memAdd);
		break;
		
		case D0_QWC:	//Ch0 Quad word count
			DMAC_LOG("%8.8lx:[DMAC][RD]D0_TADG : 0x%8.8lx ", (long)MipsCpu.PC, DmaC.Channel[ DMA_CH0 ]->size);
		break;
				
		case D0_TADG:	//Ch0 Tag address 
			DMAC_LOG("%8.8lx:[DMAC][RD]D0_TADG : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH0 ]->tag);
		break;
						
		case D0_ASR0:	//CH0 Address stack 0
			DMAC_LOG("%8.8lx:[DMAC][RD]D0_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH0 ]->asr0);
		break;
				
		case D0_ASR1:	//Ch0 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][RD]D0_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH0 ]->asr1);
		break;

/* DMA CH1 REGISTERS (Linked to VIF1) */
		case D1_CHCR: 	//Ch1 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D1_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH1 ]->ctrl);
		break;
		
		case D1_MADR: 	//Ch1 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D1_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH1 ]->memAdd);
		break;
		
		case D1_QWC:	//Ch1 Quad word count		
			DMAC_LOG("%8.8lx:[DMAC][RD]D1_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH1 ]->size);
		break;
	
		case D1_TADR:	//Ch1 Tag address			
			DMAC_LOG("%8.8lx:[DMAC][RD]D1_TADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH1 ]->tag);
		break;
		
		case D1_ASR0:	//CH1 Address stack 0			
			DMAC_LOG("%8.8lx:[DMAC][RD]D1_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH1 ]->asr0);
		break;
		
		case D1_ASR1:	//CH1 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][RD]D1_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH1 ]->asr1);
		break;

/* DMA CH2 REGISTERS (Linked to GIF) */
		case D2_CHCR: 	//Ch2 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D2_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH2 ]->ctrl);
		break;
		
		case D2_MADR: 	//Ch2 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D2_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH2 ]->memAdd);
		break;
		
		case D2_QWC:	//Ch2 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D2_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH2 ]->size);
		break;
	
		case D2_TADR:	//Ch2 Tag address			
			DMAC_LOG("%8.8lx:[DMAC][RD]D2_TADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH2 ]->tag);
		break;
		
		case D2_ASR0:	//CH2 Address stack 0			
			DMAC_LOG("%8.8lx:[DMAC][RD]D2_ASR0 : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH2 ]->asr0);
		break;
		
		case D2_ASR1:	//CH2 Address stack 1			
			DMAC_LOG("%8.8lx:[DMAC][RD]D2_ASR1 : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH2 ]->asr1);
		break;


/* DMA CH3 REGISTERS (Linked to IPU (FROM IPU)) */
		case D3_CHCR: 	//Ch3 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D3_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH3 ]->ctrl);
		break;
		
		case D3_MADR: 	//Ch3 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D3_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH3 ]->memAdd);
		break;
		
		case D3_QWC:	//Ch3 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D3_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH3 ]->size);
		break;

/* DMA CH4 REGISTERS (Linked to IPU (TO IPU)) */
		case D4_CHCR: 	//Ch4 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D4_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH4 ]->ctrl);
		break;
	
		case D4_MADR: 	//Ch4 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D4_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH4 ]->memAdd);
		break;
		
		case D4_QWC:	//Ch4 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D4_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH4 ]->size);
		break;
	
		case D4_TADR: 	//Ch4 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D4_TADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH4 ]->tag);
		break;



/* DMA CH5 REGISTERS (Linked to SIF0) */
		case D5_CHCR: 	//Ch5 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D5_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH5 ]->ctrl);
		break;
		
		case D5_MADR: 	//Ch5 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D5_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH5 ]->memAdd);
		break;
		
		case D5_QWC:	//Ch5 Quad word count		
			DMAC_LOG("%8.8lx:[DMAC][RD]D5_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH5 ]->size);
		break;


/* DMA CH6 REGISTERS (Linked to SIF1) */
		case D6_CHCR: 	//Ch6 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D6_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH6 ]->ctrl);
		break;
	
		case D6_MADR: 	//Ch6 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D6_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH6 ]->memAdd);
		break;
		
		case D6_QWC:	//Ch6 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D6_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH6 ]->size);
		break;
		
		case D6_TADR: 	//Ch6 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D6_TADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH6 ]->tag);
		break;


/* DMA CH7 REGISTERS (Linked to SIF2) */
		case D7_CHCR: 	//Ch7 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D7_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH7 ]->ctrl);
		break;
		
		case D7_MADR: 	//Ch7 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D7_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH7 ]->memAdd);
		break;
		
		case D7_QWC:	//Ch7 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D7_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH7 ]->size);
		break;


/* DMA CH8 REGISTERS (Linked to SPR  (from SCRATCH PAD) */
		case D8_CHCR: 	//Ch8 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D8_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH8 ]->ctrl);
		break;
	
		case D8_MADR: 	//Ch8 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D8_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH8 ]->memAdd);
		break;
		
		case D8_QWC:	//Ch8 Quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D8_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH8 ]->size);
		break;
		
		case D8_SADR:	//Ch8 SPR address			
			DMAC_LOG("%8.8lx:[DMAC][RD]D8_SADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH8 ]->saddr);
		break;


/* DMA CH9 REGISTERS (Linked to SPR  (To SCRATCH PAD) */
		case 0x1000D401:
		case D9_CHCR: 	//Ch9 Channel Control		
			DMAC_LOG("%8.8lx:[DMAC][RD]D9_CHCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH9 ]->ctrl);
		break;
	
		case D9_MADR: 	//Ch9 Memory Address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D9_MADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH9 ]->memAdd);
		break;
		
		case D9_QWC:	//Ch9 quad word count			
			DMAC_LOG("%8.8lx:[DMAC][RD]D9_QWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH9 ]->size);
		break;
		
		case D9_TADR: 	//Ch9 Tag address		
			DMAC_LOG("%8.8lx:[DMAC][RD]D9_TADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH9 ]->tag);
		break;
		
		case D9_SADR:	//Ch9 SPR address			
			DMAC_LOG("%8.8lx:[DMAC][RD]D9_SADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Channel[ DMA_CH9 ]->saddr);
		break;




/* DMAC REGISTERS */
		case DMAC_CTRL:   //DMAC Control           
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_CTRL : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->ctrl);
		break;
		
		case DMAC_STAT:   //DMAC Status           
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_STAT : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->stat);
		break;
		
		case DMAC_PCR:    //DMAC Priority control           
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_PCR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->pcr);
		break;
	
		case DMAC_SQWC:   //DMAC skip quad word           
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_SQWC : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->sqwc);
		break;
		
		case DMAC_RBSR:    //DMAC ring buffer size          
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_RBSR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->rbsr);
		break;
		
		case DMAC_RBOR:    //DMAC Ring buffer offset          
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_RBOR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->rbor);
		break;
		
		case DMAC_STADR:  //DMA stall address           
			DMAC_LOG("%8.8lx:[DMAC][RD]DMAC_STADR : 0x%8.8lx ", (long)MipsCpu.PC,DmaC.Reg->stadr);
		break;
		
		default:
			EMU_LOG("%8.8lx:[DMAC][RD]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
		break;
	}
#endif

	RealAdd 		= fPhysicalAddress & 0xFFFF;

	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryReadDQ 	( SpecialAreaMemory, fVal, RealAdd );break;
					
		case DATA_DW:	MipsMemoryReadDW	( SpecialAreaMemory, fVal, RealAdd );break;

	    case DATA_HW:	MipsMemoryReadHW	( SpecialAreaMemory, fVal, RealAdd );break;
		
		case DATA_W:	MipsMemoryReadW 	( SpecialAreaMemory, fVal, RealAdd );break;

		case DATA_B:	MipsMemoryReadB		( SpecialAreaMemory, fVal, RealAdd );break;

		default:		EMU_LOG("%8.8lx:[DMAC][RD][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
						MipsCpu.halt = 1;break;
	}


}














