/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>


#include "nSX2.h"
#include "PS2Edefs.h"

#include "GIF.h"


GIFGSInterface GIF;


void GIF_Log ( char * ftext  , ...   ) {
	char GIF_LogText[ FILE_NAME_LENGHT ];

	if ( debug_gif == 1 ) {

		va_list ap;
		va_start (ap, ftext);
  		vsprintf (GIF_LogText, ftext, ap);
   		va_end (ap);

		PrintToGIFOutput 		( GIF_LogText );

	}

}




void 	GIF_init( void ) {
#ifdef EMU_LOG
	EMU_LOG("[GIF]Initialisation");
#endif

	/* Emulator Initialisation */
	GIF.Reg = ( GIFRegisters * ) &SpecialAreaMemory[   ( GIF_CTRL & 0xFFFF )  ];

}



/* Write to GIF Registers */
void	GIF_Write ( _u64 fPhysicalAddress , _u128 fVal , _u8 fDataType ) {
	_u32	boucle;
	_u64	RealAdd;


	RealAdd 		= fPhysicalAddress & 0xFFFF;

	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.lo, RealAdd );
						MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.hi, RealAdd + 8 );break;

			
		case DATA_DW:	MipsMemoryWriteDW	( SpecialAreaMemory, fVal.lo, RealAdd );break;
			
		case DATA_W:	MipsMemoryWriteW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

		case DATA_HW:	MipsMemoryWriteHW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

		case DATA_B:	MipsMemoryWriteB	( SpecialAreaMemory, fVal.lo, RealAdd );break;
										
		default:		EMU_LOG("%8.8lx:[GIF][WR][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
						MipsCpu.halt = 1;break;
	}


	switch ( fPhysicalAddress  ) {

		case 	GIF_CTRL:	//Control
#ifdef GIF_LOG
			GIF_LOG("%8.8lx:[GIF][WR]GIF_CTRL : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
		
			switch ( fVal.lo ) {

				case 0x1 :
#ifdef GIF_LOG_ST
					GIF_LOG_ST("%8.8lx:[GIF][WR]GIF_CTRL : RESET ", (long)MipsCpu.PC);
#endif
					GIF.Reg->ctrl = GIF.Reg->ctrl & ~(0x1);
				break;

				default:
#ifdef GIF_LOG_ST
					GIF_LOG_ST("%8.8lx:[GIF][WR]GIF_CTRL : UNKNOW ( %2.2x ) ", (long)MipsCpu.PC,fVal.lo);
#endif
				break;

			}
		break;


		case 	GIF_MODE:	//Mode setting
#ifdef GIF_LOG_ST
			GIF_LOG_ST("%8.8lx:[GIF][WR]GIF_MODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
		break;

#ifdef GIF_LOG
		case	GIF_STAT:	//Status
			GIF_LOG("%8.8lx:[GIF][WR]GIF_STAT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_TAG0:	//GIFtag (bits 31-0) immediately before
			GIF_LOG("%8.8lx:[GIF][WR]GIF_TAG0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_TAG1:	//GIFtag (bits 63-32) immediately before
			GIF_LOG("%8.8lx:[GIF][WR]GIF_TAG1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_TAG2:	//GIFtag (bits 95-64) immediately before
			GIF_LOG("%8.8lx:[GIF][WR]GIF_TAG2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_TAG3:	//GIFtag (bits 127-96) immediately before
			GIF_LOG("%8.8lx:[GIF][WR]GIF_TAG3 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_CNT:	//Transfer status counter
			GIF_LOG("%8.8lx:[GIF][WR]GIF_CNT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_P3CNT:	//PATH3 transfer status counter
			GIF_LOG("%8.8lx:[GIF][WR]GIF_P3CNT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case	GIF_P3TAG:	//PATH3 tag value
			GIF_LOG("%8.8lx:[GIF][WR]GIF_P3TAG : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif
		default: 
			EMU_LOG("%8.8lx:[GIF][WR]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
		break;

	}
}




/* Write to GIF Registers */
void	GIF_Read ( _u64 fPhysicalAddress , _u128 * fVal , _u8 fDataType ) {

	_u32	boucle;
	_u64	RealAdd;


#ifdef GIF_LOG
	switch ( fPhysicalAddress  ) {
		case 	GIF_CTRL:	//Control
			GIF_LOG("%8.8lx:[GIF][RD]GIF_CTRL : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->ctrl);
		break;

		case 	GIF_MODE:	//Mode setting
			GIF_LOG("%8.8lx:[GIF][RD]GIF_MODE : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->mode);
		break;

		case	GIF_STAT:	//Status
			GIF_LOG("%8.8lx:[GIF][RD]GIF_STAT : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->stat);
		break;

		case	GIF_TAG0:	//GIFtag (bits 31-0) immediately before
			GIF_LOG("%8.8lx:[GIF][RD]GIF_TAG0 : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->tag0);
		break;

		case	GIF_TAG1:	//GIFtag (bits 63-32) immediately before
			GIF_LOG("%8.8lx:[GIF][RD]GIF_TAG1 : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->tag1);
		break;

		case	GIF_TAG2:	//GIFtag (bits 95-64) immediately before
			GIF_LOG("%8.8lx:[GIF][RD]GIF_TAG2 : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->tag2);
		break;

		case	GIF_TAG3:	//GIFtag (bits 127-96) immediately before
			GIF_LOG("%8.8lx:[GIF][RD]GIF_TAG3 : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->tag3);
		break;

		case	GIF_CNT:	//Transfer status counter
			GIF_LOG("%8.8lx:[GIF][RD]GIF_CNT : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->cnt);
		break;

		case	GIF_P3CNT:	//PATH3 transfer status counter
			GIF_LOG("%8.8lx:[GIF][RD]GIF_P3CNT : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->p3cnt);
		break;

		case	GIF_P3TAG:	//PATH3 tag value
			GIF_LOG("%8.8lx:[GIF][RD]GIF_P3TAG : 0x%8.8lx ", (long)MipsCpu.PC, GIF.Reg->p3tag);
		break;

		default: 
			EMU_LOG("%8.8lx:[GIF][RD]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
		break;

	}
#endif


	RealAdd 		= fPhysicalAddress & 0xFFFF;

	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryReadDQ 	( SpecialAreaMemory, fVal, RealAdd );break;
					
		case DATA_DW:	MipsMemoryReadDW	( SpecialAreaMemory, fVal, RealAdd );break;

	    case DATA_HW:	MipsMemoryReadHW	( SpecialAreaMemory, fVal, RealAdd );break;
		
		case DATA_W:	MipsMemoryReadW 	( SpecialAreaMemory, fVal, RealAdd );break;

		case DATA_B:	MipsMemoryReadB		( SpecialAreaMemory, fVal, RealAdd );break;

		default:		EMU_LOG("%8.8lx:[GIF][RD][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
						MipsCpu.halt = 1;break;
	}



}







