/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#include <stdlib.h>
#include <stdio.h>



#include "nSX2.h"
#include "PS2Edefs.h"

/* must create function to avoid this kind of declaration */
#include "IntrC.h"


void GS_Log ( char * ftext  , ...   ) {
	char GS_LogText[ FILE_NAME_LENGHT ];

	if ( debug_gs == 1 ) {

		va_list ap;
		va_start (ap, ftext);
	  	vsprintf (GS_LogText, ftext, ap);
	   	va_end (ap);

		sprintf(GS_LogText,"%s\n",GS_LogText);

		PrintToGSOutput( GS_LogText );
	}
//	PrintToEmuOutput (  GS_LogText ); 


}



/* Write to GS Registers */
void	GS_Write ( _u64 fPhysicalAddress , _u128 fVal , _u8 fDataType ) {
	_u32	boucle;
	_u64	RealAdd;


	RealAdd 		= fPhysicalAddress & 0x1FFF;

	switch ( fDataType ) {
			
		case DATA_DW:	MipsMemoryWriteDW	( GSAreaMemory, fVal.lo, RealAdd );
						(*GSwrite64)	( RealAdd, (_u64)fVal.lo );
						break;

		case DATA_W:	MipsMemoryWriteW	( GSAreaMemory, fVal.lo, RealAdd );
						(*GSwrite32)	( RealAdd, (_u32)fVal.lo);
						break;
										
		default:		EMU_LOG("%8.8lx:[GS][WR][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
						MipsCpu.halt = 1;break;
	}

#ifdef GS_LOG
	switch ( fPhysicalAddress  ) {

		case GS_PMODE:		//Various PCRTC modes
			GS_LOG("%8.8lx:[GS][WR]GS_PMODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_SMODE1:		//Related to Sync
			GS_LOG("%8.8lx:[GS][WR]GS_SMODE1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_SMODE2:		//Related to Sync
			GS_LOG("%8.8lx:[GS][WR]GS_SMODE2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_SRFSH:		//DRAM refresh
			GS_LOG("%8.8lx:[GS][WR]GS_SRFSH : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_SYNCH1:		//Related to Sync
			GS_LOG("%8.8lx:[GS][WR]GS_SYNCH1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_SYNCH2:		//Related to Sync
			GS_LOG("%8.8lx:[GS][WR]GS_SYNCH2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_SYNCV:		//Related to Sync/start
			GS_LOG("%8.8lx:[GS][WR]GS_SYNCV : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_DISPFB1:		//Related to display buffer of Rectangular Area 1
			GS_LOG("%8.8lx:[GS][WR]GS_DISPFB1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_DISPLAY1:		//Rectangular Area 1 display position etc.
			GS_LOG("%8.8lx:[GS][WR]GS_DISPLAY1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_DISPFB2:		//Related to display buffer of Rectangular Area 2
			GS_LOG("%8.8lx:[GS][WR]GS_DISPFB2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_DISPLAY2:		//Rectangular Area 2 display position etc.
			GS_LOG("%8.8lx:[GS][WR]GS_DISPLAY2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_EXTBUF:		//Rectangular Area write buffer
			GS_LOG("%8.8lx:[GS][WR]GS_EXTBUF : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_EXTDATA:		//Rectangular Area write data
			GS_LOG("%8.8lx:[GS][WR]GS_EXTDATA : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_EXTWRITE:		//Rectangular Area write start
			GS_LOG("%8.8lx:[GS][WR]GS_EXTWRITE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_BGCOLOR:		//Background color
			GS_LOG("%8.8lx:[GS][WR]GS_BGCOLOR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
		case GS_IMR	:		///IMR
			GS_LOG("%8.8lx:[GS][WR]GS_IMR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		default: 
			EMU_LOG("%8.8lx:[GS][WR]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
		break;

	}
#endif

}


void GSUpdate( void ) {
	static	_u8	Toggle_VBLANK = 1;
 	
	/*
	if need refresh Screen
	This is done when Reading at 0x12001000 ( r5900_memory.c ) or
	when Interrupt occurs on Counter 4 ( timers_main.c )
	*/	
	if ( Ps2Bios.NeedUpdateScreen == TRUE ) {
		
		(*GSvsync)();

		
		if ( Toggle_VBLANK == 1 ) {

			INTC_RaiseIntByCause ( INTC_VBLANK_E );
			Toggle_VBLANK = 0;
		} else {
			INTC_RaiseIntByCause ( INTC_VBLANK_S );
			Toggle_VBLANK = 1;
		}

		/* I don't need to refresh screeen */
		Ps2Bios.NeedUpdateScreen = FALSE;

	}

}







