/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include "nSX2.h"


_u8	* MipsMemory 		= NULL;
_u8	* SpecialAreaMemory = NULL;
_u8	* GSAreaMemory 		= NULL;
_u8 * BiosMemory 		= NULL;
_u8 * SPRMemory			= NULL;


/****************************************************************************************************/


/*
0x00100000-0x01ffffff RAM Mirror Address. (cached)
0x20100000-0x21ffffff RAM Mirror Adress. (not cached)
0x30100000-0x31ffffff RAM Mirror Adress. (not cached & accelerated)
*/
_u8 MipsMemoryAlloc( void )  {
	_u8 	Result;
	
	
	MipsMemory = malloc ( sizeof( _u8 ) * 0x01FFFFFF );
	if (MipsMemory != NULL ) {
		memset( MipsMemory, 0, sizeof( _u8 ) * 0x01FFFFFF );
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc Main 32 Mbits : OK");
#endif
		Result = TRUE;
	}
	else {
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc Main 32 Mbits : Failed");
#endif
		Result = FALSE;
	}
	


	return Result;
}


/*
0x10000000 - 0x1000ffff

COUNTERS' REGISTERS
IPU REGISTERS (MPEG2 DECODER)
GIF REGISTERS
VIF0 REGISTERS
VIF1 REGISTERS
FIFO
DMA CH0 REGISTERS (Linked to VIF0)
DMA CH1 REGISTERS (Linked to VIF1)
DMA CH2 REGISTERS (Linked to GIF)
DMA CH3 REGISTERS (Linked to IPU (FROM???))
DMA CH4 REGISTERS (Linked to IPU (TO???))
DMA CH5 REGISTERS (Linked to SIF0)
DMA CH6 REGISTERS (Linked to SIF1)
DMA CH7 REGISTERS (Linked to SIF2)
DMA CH8 REGISTERS (Linked to SPR  (form SCRATCH PAD to RAM???)
DMA CH9 REGISTERS (Linked to SPR  (form RAM to SCRATCH PAD???)
DMA CONTROL REGISTERS
*/
_u8 MipsSpecialAreaMemoryAlloc( void )  {
	_u8 	Result;	

    SpecialAreaMemory	 = NULL;
	SpecialAreaMemory = malloc ( sizeof( _u8 ) * 0xFFFF );
	if (SpecialAreaMemory != NULL ) {
		memset( SpecialAreaMemory, 0, sizeof( _u8 ) * 0xFFFF );
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc Special 64 kbytes : OK");
#endif

		Result = TRUE;
	}
	else {
		
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc Special 64 kbytes : Failed");
#endif
		Result = FALSE;
	}



	return Result;
}

/*
0x12000000 - 0x120010C0
*/
_u8 MipsGSAreaMemoryAlloc( void )  {
	_u8 	Result;	

	
    GSAreaMemory = NULL;
	GSAreaMemory = malloc ( sizeof( _u8 ) * 0x1FFF );
	if (GSAreaMemory != NULL ) {
  		memset( GSAreaMemory, 0, sizeof( _u8 ) * 0x1FFF );
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc GS 4288 bytes : OK");
#endif
		Result = TRUE;
	}
	else {
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc GS 4288 bytes : Failed");
#endif
		Result = FALSE;
	}
	

	
	return Result;
}



/*
0x1FC00000 - 0x1FFFFFFF ROM BIOS Mirror Adress. (not cached)
0x9FC00000 - 0x9FFFFFFF ROM BIOS Mirror Adress. (cached)
0xBFC00000 - 0xBFFFFFFF ROM BIOS Mirror Adress. (not cached)
*/
_u8 MipsBiosMemoryAlloc( void )  {
	_u8 	Result;	
	
	BiosMemory = malloc ( sizeof( _u8 ) * 0x3FFFFF );
	if (BiosMemory != NULL ) {
		memset( BiosMemory, 0, sizeof( _u8 ) * 0x3FFFFF );
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc BIOS 4 Mbytes : OK");
#endif
		Result = TRUE;
	}
	else {
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc BIOS 4 Mbytes : Failed");
#endif
		Result = FALSE;
	}
	

	
	return Result;
}

_u8 MipsSPRMemoryAlloc( void )  {
	_u8 	Result;	

	
	SPRMemory = malloc ( sizeof( _u8 ) * 0x3FFF );
	if (SPRMemory != NULL ) {
		memset( SPRMemory, 0, sizeof( _u8 ) * 0x3FFF );
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc SPR 16 Kbytes : OK");
#endif
		Result = TRUE;
	}
	else {
#ifdef EMU_LOG
		EMU_LOG("[MEM_MAN]Alloc SPR 16 Kbytes : Failed");
#endif
		Result = FALSE;
	}
	

	
	return Result;
}



void FreeAllMemory( void ) {
	char tempText[100];
	
	if ( MipsMemory != NULL ) {			free(MipsMemory );			MipsMemory 			= NULL; }
	if ( SpecialAreaMemory != NULL ) {	free(SpecialAreaMemory);	SpecialAreaMemory 	= NULL; }
	if ( GSAreaMemory != NULL ) {		free(GSAreaMemory);			GSAreaMemory		= NULL; }
	if ( BiosMemory != NULL ) 	{		free(BiosMemory);			BiosMemory 			= NULL; }
	if ( SPRMemory != NULL ) 	{		free(SPRMemory);			SPRMemory 			= NULL; }

#ifdef EMU_LOG
	EMU_LOG("[MEM_MAN]Free Memory");
#endif
	
}



