/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>




#include "nSX2.h"
#include "PS2Edefs.h"

/*
Read Values from Pluggin
*/
void PadReadStatus(int pad) {
	char tempString[256];

	int i = pad-1, j;
	int bufcount;
	u8 pad_buf[32];


	memset(pad_buf,0,32);

	/* If Pad Data is alrealdy Initialised - done in RPC_8000010f.c*/
	if (Ps2Bios.PadState[i] != NULL) {
		/* The controller is connected and Request is done */
		if (Ps2Bios.PadData[i][0] != NULL) {
			Ps2Bios.PadData[i][0]->state = PAD_STATE_STABLE;
			Ps2Bios.PadData[i][0]->reqState = PAD_RSTAT_COMPLETE;
		}


		/* The controller is connected and Request is done */
		if (Ps2Bios.XPadData[i][0] != NULL) {
			Ps2Bios.XPadData[i][0]->state = PAD_STATE_STABLE;
			Ps2Bios.XPadData[i][0]->reqState = PAD_RSTAT_COMPLETE;
		}


	/* Start Poll on pad - Data */
	PAD1startPoll(pad);
	
	/* Get data[0] from pluggin  */	
	pad_buf[0] = 0; 
	pad_buf[1] = PAD1poll(/*0x42*/0); 
	
	/*
	0 - 7 : DATA
	*/
	/* ?? */
	if (!(pad_buf[1] & 0x0f)) { 
		bufcount = 32; 
	} else { 
		bufcount = (pad_buf[1] & 0x0f) * 2; 
	} 
	/* do no take data[1] from pluggin*/
	PAD1poll(0); 
	
	/* Get data from data[2] -> data[2+buffcount] */
	j = 2; 
	while (bufcount--) { 
		pad_buf[j++] = PAD1poll(0); 
	} 


	/* Copy Pad value 
		Only : Ok - btns[0] & [1] 
		Ananlog will come after
		Need To Update the Input Plugin Before, ( to Get the all 32 bytes )
	*/
	
	Ps2Bios.PadState[i]->ok = 0;
	
	/* Decode the Struct */

	Ps2Bios.PadState[i]->btns[0] = pad_buf[2]; /* data[2] */
	Ps2Bios.PadState[i]->btns[1] = pad_buf[3]; /* data[3] */

/*
	sprintf(tempString,"DataRead: 0:%x 1:%x \n",Ps2Bios.PadState[i]->btns[0],Ps2Bios.PadState[i]->btns[1]);
	PrintToEmuOutput (  tempString );
*/
	
	/*  Hack of Right Analogic Joystick ( for turnip )
		Copy button value to Analogic Joystick 
	*/
	switch ( pad_buf[2] ) {	
	
		// Up 
		case 0xef:
			if ( Ps2Bios.PadState[i]->rjoy_v == 0x00 ) Ps2Bios.PadState[i]->rjoy_v = 0x1;
			Ps2Bios.PadState[i]->rjoy_v = Ps2Bios.PadState[i]->rjoy_v - 1;

			
		break;
		// down 
		case 0xbf:

			if ( Ps2Bios.PadState[i]->rjoy_v == 0xFF ) Ps2Bios.PadState[i]->rjoy_v = 0xFE;
			Ps2Bios.PadState[i]->rjoy_v = Ps2Bios.PadState[i]->rjoy_v + 1;

			



		break;

		
		default:  

		break;

		}

	}
}



/*
Handle Key Event from GS Pluggin
*/
void KeyEvent(keyEvent* ev) {
	char tempString[256];

	/* If no Event , Exit */
	if (ev == NULL) return;

	/* Handle Event only on KeyRealease */
	if (ev->event == !KEYRELEASE) return;

	switch (ev->key) {
		
		/* F8 Key */
		case 0x77:
			(*GSmakeSnapshot)("screenshots\\");
		break;

/* Doesn't seems to work very good... */
/*
		case 0x76: // F7 Key - Pause
		    MipsCpu.halt = 1;
#ifdef EMU_LOG
			EMU_LOG("Pause");
#endif		
    		break;

		case 0x75: // F6 Key - Run
 	        MipsCpu.halt = 0;
#ifdef EMU_LOG
			EMU_LOG("Run");
#endif		
   	        break;
*/
 	    case 0x1b: /* Esc Key */
           	GUI_SendStopMainLoop( );
         break;


		default:
			//sprintf(tempString,"KeyPressed:%8.8lx\n",ev->key);PrintToEmuOutput (  tempString );
			(*GSkeyEvent)(ev);		
		break;
	}
}







void PadAndKeyUpdate( void ) {

	/* Get Event from Pluggin and pass it to KeyEvent Func */
	
	KeyEvent( (*PAD1keyEvent)() );	

	/* If need to Read Pad Key - Set in timers_main.c */
	if ( Ps2Bios.NeedUpdatePad == TRUE ) {

		// If Pad is already Initialised 
		if ( Ps2Bios.PadState[0] != NULL ) { 
			PadReadStatus(1);
			//PadReadStatus(2);
		}
	
		/* Don't need to update the next time */
		Ps2Bios.NeedUpdatePad = FALSE;

	}


}

