/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include "nSX2.h"
#include "PS2Edefs.h"

#include "timers.h"

#include "Timers_cycles.h"



/* Get the number of Clock Cycles for specified Opcodes */
_u8 Timers_GetCycles	( _u32 fOpcode ) {

	_u8		ReturnValue = 0;
	

	switch ( ((fOpcode >> 26) & 0x3F) ) {
	
		case 0 	: 	ReturnValue = special_cycles	[ ( fOpcode & 0x3F)				]; break;
		case 1	: 	ReturnValue = regimm_cycles		[ ( fOpcode & 0x1F0000 )>> 16	]; break;
		case 16 : 	ReturnValue = cop0_cycles		[ ( fOpcode & 0x3E00000)>> 21	]; break;
		case 17 : {
					switch( ( fOpcode & 0x3E00000 ) >> 21 ) {
						
						case 8 : ReturnValue = cop1_bc1_cycles	[ ( fOpcode >> 16)& 0x3 ]; break;					
						case 16: ReturnValue = cop1_s_cycles	[   fOpcode       & 0x3F ]; break;					
						case 20: ReturnValue = cop1_WL_cycles	[   fOpcode       & 0x3F ]; break;					

						default: ReturnValue = cop1_cycles		[ ( fOpcode&0x3E00000)>>21];break;					
					}
					
					break;
				}	
		case 18	: {
					if ( ((fOpcode&0x3E00000)>>21) > 15 ) {
						switch ( fOpcode&0x3F ) {

							case 60 : 	ReturnValue = cop2_V60_cycles	[ ( fOpcode & 0x7C0 ) >> 6]; break;
							case 61 : 	ReturnValue = cop2_V61_cycles	[ ( fOpcode & 0x7C0 ) >> 6]; break;
							case 62 : 	ReturnValue = cop2_V62_cycles	[ ( fOpcode & 0x7C0 ) >> 6]; break;
							case 63 : 	ReturnValue = cop2_V63_cycles	[ ( fOpcode & 0x7C0 ) >> 6]; break;
							default : 	ReturnValue = cop2_V_cycles		[   fOpcode & 0x3F        ];break;
						}
					} else				ReturnValue = cop2_cycles		[ ( fOpcode & 0x3E00000)>>21];
				

				}
				break;

		case 28 : {
				switch  (MipsCpu.opcode  & 0x3F) {
			
				case 8:		ReturnValue = MMI0_cycles[ ( ( fOpcode >> 6 )  & 0x1F )]; break;
				case 9:		ReturnValue = MMI1_cycles[ ( ( fOpcode >> 6 )  & 0x1F )]; break;
				case 40:	ReturnValue = MMI2_cycles[ ( ( fOpcode >> 6 )  & 0x1F )]; break;
				case 41:	ReturnValue = MMI3_cycles[ ( ( fOpcode >> 6 )  & 0x1F )]; break;
				
				default : 	ReturnValue = special2_cycles[ (  fOpcode   & 0x3F)] ;
				}			

			}
			break;
	
		default : ReturnValue = normal_cycles[((fOpcode >> 26) & 0x3F)]; break;
	
	}

	return ReturnValue;

}

