/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#include "nSX2_Type.h"
 

/* CORE */
_u8 normal_cycles[64] =
{
	0/*SPECIAL*/,	0/*REGIMM*/,	6/*J*/,		6/*JAL*/,	6/*BEQ*/,		6/*BNE*/,	6/*BLEZ*/,	6/*BGTZ*/,
	4/*ADDI*/, 		4/*ADDIU*/, 	4/*SLTI*/,	4/*SLTIU*/,	4/*ANDI*/,		4/*ORI*/,	4/*XORI*/,	4/*LUI*/,
	0/*COP0*/, 		0/*COP1*/, 		0/*COP2*/,	0/*NI*/,	6/*BEQL*/,		6/*BNEL*/,	6/*BLEZL*/,	6/*BGTZL*/,
	4/*DADDI*/, 	4/*DADDIU*/, 	6/*LDL*/,	6/*LDR*/,	0/*SPECIAL2*/,	0/*NI*/,	6/*LQ*/,	6/*SQ*/,
	6/*LB*/, 		6/*LH*/, 		6/*LWL*/,	6/*LW*/,	6/*LBU*/,		6/*LHU*/,	6/*LWR*/,	6/*LWU*/,
	6/*SB*/, 		6/*SH*/, 		6/*SWL*/,	6/*SW*/,	6/*SDL*/,		6/*SDR*/,	6/*SWR*/,	2/*CACHE*/,
	0/*NI*/, 		6/*LWC1*/, 		0/*NI*/,	2/*PREF*/,	0/*NI*/,		0/*NI*/,	6/*LQC2*/,	6/*LD*/,
	0/*NI*/, 		6/*SWC1*/, 		0/*NI*/,	0/*NI*/,	0/*NI*/, 		0/*NI*/,	6/*SQC2*/,	6/*SD*/
};




_u8  special_cycles[64] =
{
	4/*SLL*/,	0/*NI*/,	4/*SRL*/,	4/*SRA*/,	4/*SLLV*/,		0/*NI*/,		4/*SRLV*/,	4/*SRAV*/,
	6/*JR*/,	6/*JALR*/,	4/*MOVZ*/,	4/*MOVN*/,	128/*SYSCALL*/,	2/*BREAK*/,		0/*NI*/,	2/*SYNC*/,
   	2/*MFHI*/,	2/*MTHI*/,	2/*MFLO*/,	2/*MTLO*/,	4/*DSLLV*/,		0/*NI*/,		4/*DSRLV*/,	4/*DSRAV*/,
   	4/*MULT*/,	4/*MULTU*/,	29/*DIV*/,	29/*DIVU*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,	0/*NI*/,
 	4/*ADD*/,	4/*ADDU*/,	4/*SUB*/,	4/*SUBU*/,	4/*AND*/,		4/*OR*/,		4/*XOR*/,	4/*NOR*/,
   	2/*MFSA*/,	2/*MTSA*/,	4/*SLT*/,	4/*SLTU*/,	4/*DADD*/,		4/*DADDU*/,		4/*DSUB*/,	4/*DSUBU*/,
   	6/*TGE*/,	6/*TGEU*/,	6/*TLT*/,	6/*TLTU*/,	6/*TEQ*/,		0/*NI*/,		0/*TNE*/,	0/*NI*/,
  	4/*DSLL*/,	0/*NI*/,	4/*DSRL*/,	4/*DSRA*/,	4/*DSLL32*/,	0/*NI*/,		4/*DSRL32*/,4/*DSRA32*/
};


_u8 regimm_cycles[32] =
{
   6/*BLTZ*/,	6/*BGEZ*/,		6/*BLTZL*/,		6/*BGEZL*/,		0/*NI*/,		0/*NI*/,	0/*NI*/,		0/*NI*/,
   6/*TGEI*/,	6/*TGEIU*/, 	6/*TLTI*/,		6/*TLTIU*/,		6/*TEQI*/,		0/*NI*/,	6/*TNEI*/,		0/*NI*/,
   6/*BLTZAL*/,	6/*BGEZAL*/, 	6/*BLTZALL*/,	6/*BGEZALL*/,	0/*NI*/,		0/*NI*/,	0/*NI*/,		0/*NI*/,
   4/*MTSAB*/,	4/*MTSAH*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/, 	0/*NI*/,		0/*NI*/ 
};


/* MMI */

_u8 special2_cycles[64] =
{
	4/*MADD*/,	4/*MADDU*/,	0/*NI*/,		0/*NI*/,		4/*PLZCW*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,	
	0/*MMI0*/,	0/*MMI2*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,	
	2/*MFHI1*/,	2/*MTHI1*/,	2/*MFLO1*/,		2/*MTLO1*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,	
	4/*MULT1*/,	4/*MULTU1*/,37/*DIV1*/,		37/*DIVU1*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,	
	4/*MADD1*/,	4/*MADDU1*/,0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,	
	0/*MMI1*/,	0/*MMI3*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,	
	4/*PMFHL*/,	4/*PMTHL*/,	0/*NI*/,		0/*NI*/,		4/*PSLLH*/,		0/*NI*/,		4/*PSRLH*/,		4/*PSRAH*/,	
	0/*NI*/,	0/*NI*/,	0/*NI*/,		0/*NI*/,		4/*PSLLW*/,		0/*NI*/,		4/*PSRLW*/,		4/*PSRAW*/							
};



_u8 MMI0_cycles[32] =
{ 
 4/*PADDW*/,	4/*PSUBW*/,		4/*PCGTW*/,		4/*PMAXW*/,		4/*PADDH*/,		4/*PSUBH*/,		4/*PCGTH*/,		4/*PMAXH*/,        
 4/*PADDB*/,	4/*PSUBB*/,		4/*PCGTB*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,
 4/*PADDSW*/,	4/*PSUBSW*/,	4/*PEXTLW*/,	4/*PPACW*/,		4/*PADDSH*/,	4/*PSUBSH*/,	4/*PEXTLH*/,	4/*PPACH*/,        
 4/*PADDSB*/,	4/*PSUBSB*/,	4/*PEXTLB*/,	4/*PPACB*/,		0/*NI*/,		0/*NI*/,		4/*PEXT5*/,		4/*PPAC5*/       
};



_u8 MMI1_cycles[32] =
{ 
 0/*NI*/,   	4/*PABSW*/,		4/*PCEQW*/,		4/*PMINW*/,		4/*PADSBH*/, 	4/*PABSH*/,  	4/*PCEQH*/, 	4/*PMINH*/, 
 0/*NI*/,   	0/*NI*/,   		4/*PCEQB*/,  	0/*NI*/,		0/*NI*/,   		0/*NI*/,   		0/*NI*/,   		0/*NI*/, 
 4/*PADDUW*/,   4/*PSUBUW*/,	4/*PEXTUW*/, 	0/*NI*/,   		4/*PADDUH*/, 	4/*PSUBUH*/, 	4/*PEXTUH*/, 	0/*NI*/, 
 4/*PADDUB*/,   4/*PSUBUB*/, 	4/*PEXTUB*/, 	4/*QFSRV*/, 	0/*NI*/,   		0/*NI*/,   		0/*NI*/,   		0/*NI*/
};


_u8 MMI2_cycles[32] = 
{ 
 4/*PMADDW*/,	0/*NI*/,		4/*PSLLVW*/,	4/*PSRLVW*/,	4/*PMSUBW*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,
 4/*PMFHI*/,	4/*PMFLO*/,		4/*PINTH*/,		0/*NI*/,		4/*PMULTW*/,	37/*PDIVW*/,	4/*PCPYLD*/,	0/*NI*/,
 4/*PMADDH*/,	4/*PHMADH*/,	4/*PAND*/,		4/*PXOR*/,		4/*PMSUBH*/,	4/*PHMSBH*/,	0/*NI*/,		0/*NI*/, 
 0/*NI*/,		0/*NI*/,		4/*PEXEH*/,		4/*PREVH*/,		4/*PMULTH*/,	4/*PDIVBW*/,	4/*PEXEW*/,		4/*PROT3W*/
};



_u8 MMI3_cycles[32] = 
{ 
 4/*PMADDUW*/,	0/*NI*/,		0/*NI*/,		4/*PSRAVW*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,
 4/*PMTHI*/,	4/*PMTLO*/,		4/*PINTEH*/,	0/*NI*/,		4/*PMULTUW*/,	37/*PDIVUW*/,	4/*PCPYUD*/,	0/*NI*/,
 0/*NI*/,		0/*NI*/,		4/*POR*/,		4/*PNOR*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,
 0/*NI*/,		0/*NI*/,		4/*PEXCH*/,		4/*PCPYH*/,		0/*NI*/,		0/*NI*/,		4/*PEXCW*/,		0/*NI*/
};




/* COP0 */

_u8 cop0_cycles[32] =
{
	2/*MF0*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,	2/*MT0*/,	0/*NI*/,		0/*NI*/,	0/*NI*/,
   	0/*BC0*/, 	0/*NI*/, 	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,		0/*NI*/, 	0/*NI*/,
  	0/*C0*/,	0/*C0*/,	0/*C0*/,	0/*C0*/,	0/*C0*/,	0/*C0*/,		0/*C0*/,	0/*C0*/,
	0/*C0*/,	0/*C0*/,	0/*C0*/,	0/*C0*/,	0/*C0*/,	0/*C0*/,		0/*C0*/,	0/*C0*/
};



_u8 cop0_c0_cycles[64] =
{
	0/*NI*/, 	4/*TLBR*/, 	4/*TLBWI*/, 0/*NI*/, 	0/*NI*/,	0/*NI*/,	4/*TLBWR*/,		0/*NI*/,
	4/*TLBP*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,		0/*NI*/,		
	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/ ,
	4/*ERET*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/ ,	 
 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/ ,
 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/ ,
 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/ ,
	2/*EI*/, 	2/*DI*/, 	0/*NI*/, 	0/*NI*/,	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/  	 
};

_u8 cop0_bc0_cycles[4] =
{
	4/*BC0F*/, 	4/*BC0T*/, 	4/*BC0FL*/, 	4/*BC0TL*/ 
};



/* COP1 */
_u8 cop1_cycles[32] =
{
	2/*MFC1*/,	0/*NI*/,	2/*CFC1*/,	0/*NI*/,	2/*MTC1*/,	0/*NI*/,	2/*CTC1*/,	0/*NI*/,
	0/*BC1*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,
	0/*COP1_S*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*COP1_W*/,0/*NI*/,	0/*NI*/, 	0/*NI*/,
	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/							
};



_u8 cop1_bc1_cycles[4] =
{
	4/*BC1F*/, 4/*BC1T*/,4/*BC1FL*/,4/*BC1TL*/
};

_u8 cop1_s_cycles[64] =
{
	4/*ADD_S*/, 4/*SUB_S*/, 4/*MUL_S*/, 17/*DIV_S*/,17/*SQRT_S*/, 	4/*ABS_S*/, 	4/*MOV_S*/, 	4/*NEG_S*/,
	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/,
	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,	0/*NI*/,		0/*NI*/,		17/*RSQRT_S*/,  0/*NI*/,
	4/*ADDA_S*/,4/*SUBA_S*/,4/*MULA_S*/,0/*NI*/, 	4/*MADD_S*/,	4/*MSUB_S*/,	4/*MADDA_S*/,	4/*MSUBA_S*/,		
	0/*NI*/,	0/*NI*/, 	0/*NI*/, 	0/*NI*/,	4/*CVT_W_S*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,
	5/*MAX_S*/, 5/*MIN_S*/, 0/*NI*/, 	0/*NI*/, 	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/,		
	2/*C_F_S*/, 0/*NI*/,	2/*C_EQ_S*/,0/*NI*/	,	2/*C_OLT_S*/,	0/*NI*/,		2/*C_OLE_S*/,	0/*NI*/,
	0/*NI*/,	0/*NI*/, 	0/*NI*/,	0/*NI*/,	0/*NI*/,		0/*NI*/,		0/*NI*/,		0/*NI*/
};



_u8 cop1_WL_cycles[64] =
{
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,		
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,
	4/*CVTS_W*/,0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,	
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,
	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/, 	0/*NI*/, 	0/*NI*/, 	0/*NI*/,0/*NI*/,
};



/* COP2 */

_u8 cop2_cycles[32] = {
	0/*NI*/, 		2/*cop2_QMFC2*/,2/*cop2_CFC2*/,0/*NI*/, 		0/*NI*/, 		2/*cop2_QMTC2*/,2/*cop2_CTC2*/, 0/*NI*/,		
	4/*cop2_BC*/, 	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/,		
	0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/,		
	0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/, 0/*cop2_Vopc*/		

};


_u8 cop2_V_cycles[64] = {

	4/*VADDbc*/,	4/*VADDbc*/,	4/*VADDbc*/,	4/*VADDbc*/,	4/*VSUBbc*/,	4/*VSUBbc*/,	4/*VSUBbc*/,	4/*VSUBbc*/,
	4/*VMADDbc*/,	4/*VMADDbc*/,	4/*VMADDbc*/,	4/*VMADDbc*/,	4/*VMSUBbc*/,	4/*VMSUBbc*/,	4/*VMSUBbc*/,	4/*VMSUBbc*/,

	4/*VMAXbc*/,	4/*VMAXbc*/,	4/*VMAXbc*/,	4/*VMAXbc*/,	4/*VMINIbc*/,	4/*VMINIbc*/,	4/*VMINIbc*/,	4/*VMINIbc*/,
	4/*VMULbc*/,	4/*VMULbc*/,	4/*VMULbc*/,	4/*VMULbc*/,	4/*VMULq*/,		4/*VMAXi*/,		4/*VMULi*/,		4/*VMINIi*/,

	4/*VADDq*/,		4/*VMADDq*/,	4/*VADDi*/,		4/*VMADDi*/,	4/*VSUBq*/,		4/*VMSUBq*/,	4/*VSUBi*/,		4/*VMSUBi*/,
	4/*VADD*/,		4/*VMADD*/,		4/*VMUL*/,		4/*VMAX*/,		4/*VSUB*/,		4/*VMSUB*/,		4/*VOPMSUB*/,	4/*VMINI*/,

	1/*VIADD*/,		1/*VISUB*/,		1/*VIADDI*/,	0/*NI*/,		1/*VIAND*/,		1/*VIOR*/,		0/*NI*/,		0/*NI*/,
	0/*VCALLMS*/,	0/*VCALLMSR*/,	0/*NI*/,		0/*NI*/,		0/*Vopc_t60*/,	0/*Vopc_t61*/,	0/*Vopc_t62*/,	0/*Vopc_t63*/

};


_u8 cop2_V60_cycles[32] = {

	4/*VADDAbc*/,	4/*VSUBAbc*/,	4/*VMADDAbc*/,	4/*VMSUBAbc*/,	4/*VITOF0*/,	4/*VFTOI0*/,	4/*VMULAbc*/,	4/*VMULAq*/,
	4/*VADDAq*/,	4/*VSUBAq*/,	4/*VADDA*/,		4/*VSUBA*/,		4/*VMOVE*/,		4/*VLQI*/,		7/*VDIV*/,		1/*VMTIR*/,
	4/*VRNEXT*/,	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/,		
	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/		

};


_u8 cop2_V61_cycles[32] = {

	4/*VADDAbc*/,	4/*VSUBAbc*/,	4/*VMADDAbc*/,	4/*VMSUBAbc*/,	4/*VITOF4*/,	4/*VFTOI4*/,	4/*VMULAbc*/,	4/*VABS*/,
	4/*VMADDAq*/,	4/*VMSUBAq*/,	4/*VMADDA*/,	4/*VMSUBA*/,	4/*VMR32*/,		4/*VSQI*/,		7/*VSQRT*/,		4/*VMFIR*/, 
	4/*VRGET*/,		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/,		
	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/		

};


_u8 cop2_V62_cycles[32] = {

	4/*VADDAbc*/,	4/*VSUBAbc*/,	4/*VMADDAbc*/,	4/*VMSUBAbc*/,	4/*VITOF12*/,	4/*VFTOI12*/,	4/*VMULAbc*/,	4/*VMULAi*/,
	4/*VADDAi*/,	4/*VSUBAi*/,	4/*VMULA*/,		4/*VOPMULA*/,	0/*NI*/,		4/*VLQD*/,		13/*VRSQRT*/,	4/*VILWR*/,
	1/*VRINIT*/,	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/,		
	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/		



};

_u8 cop2_V63_cycles[32] = {

	4/*VADDAbc*/,	4/*VSUBAbc*/,	4/*VMADDAbc*/,	4/*VMSUBAbc*/,	4/*VITOF15*/,	4/*VFTOI15*/,	4/*VMULAbc*/,	4/*VCLIP*/,
	4/*VMADDAi*/,	4/*VMSUBAi*/,	0/*NI*/,		4/*VNOP*/,		0/*NI*/,		4/*VSQD*/,		13/*VWAITQ*/,	4/*VISWR*/,
	1/*VRXOR*/,		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/,		
	0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/, 		0/*NI*/		

};
















