/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include <math.h>
#include <string.h>


#include "nSX2.h"
#include "PS2Edefs.h"

#include "VU.h"
#include "VU1_micromode_tables.h"

#include "VIF.h"

 

/* Write to VIF0 Registers */
void	VIF0_Write ( _u64 fPhysicalAddress , _u128 fVal , _u8 fDataType ) {

	_u32	boucle;
	_u64	RealAdd;



		RealAdd 		= fPhysicalAddress & 0xFFFF;

		switch ( fDataType ) {

			case DATA_QUAD:	MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.lo, RealAdd );
							MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.hi, RealAdd + 8 );break;

				
			case DATA_DW:	MipsMemoryWriteDW	( SpecialAreaMemory, fVal.lo, RealAdd );break;
				
			case DATA_W:	MipsMemoryWriteW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

			case DATA_HW:	MipsMemoryWriteHW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

			case DATA_B:	MipsMemoryWriteB	( SpecialAreaMemory, fVal.lo, RealAdd );break;
										
			default:		EMU_LOG("%8.8lx:[VIF0][WR][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
							MipsCpu.halt = 1;break;
		}





	switch ( fPhysicalAddress  ) {

		case VIF0_STAT_REG: 	  //Status
#ifdef VIF_LOG
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_STAT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif


		break;

		case VIF0_FBRST_REG: 	  	//Operation Control
#ifdef VIF_LOG
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_FBRST : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			switch ( VIF[ VIF0 ].Reg->fbrst ) {

				case 0x1: 
#ifdef VIF_LOG_L2
					VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_FBRST : RESET ", (long)MipsCpu.PC);
#endif
					VIF[ VIF0 ].Reg->fbrst = VIF[ VIF0 ].Reg->fbrst & ~(0x1);
				break;

				default:
#ifdef VIF_LOG_L2
					VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_FBRST : UNKNOW ( %2.2x ) ", (long)MipsCpu.PC,fVal.lo);
#endif
				break;
			}

		break;

		case VIF0_ERR_REG: 	  	//Error status
#ifdef VIF_LOG
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_ERR : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

			VIF[ VIF0 ].iBitInterruptEnable 		= (~(VIF[ VIF0 ].Reg->err)&0x1);
			VIF[ VIF0 ].IgnoreMismatchDmaTAGError	= VIF[ VIF0 ].Reg->err & 0x2;
			VIF[ VIF0 ].IgnoreVIFCodeError			= VIF[ VIF0 ].Reg->err & 0x4;

#ifdef VIF_LOG_L2
			if ( VIF[ VIF0 ].iBitInterruptEnable ) 
				VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_ERR : ibit Enable ", (long)MipsCpu.PC);
			else
				VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_ERR : ibit Disable ", (long)MipsCpu.PC);

			if ( VIF[ VIF0 ].IgnoreMismatchDmaTAGError ) 
				VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_ERR : Ignore DMATAG ERROR ", (long)MipsCpu.PC);
			else
				VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_ERR : stall when  DMATAG ERROR ", (long)MipsCpu.PC);

			if ( VIF[ VIF0 ].IgnoreVIFCodeError ) 
				VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_ERR : Ignore VIFCode ERROR (as NOP) ", (long)MipsCpu.PC);
			else
				VIF_LOG_L2("%8.8lx:[VIF0][WR]VIF0_ERR : stall when VIFCode ERROR ", (long)MipsCpu.PC);
#endif
		break;

#ifdef VIF_LOG
		case VIF0_MARK_REG: 	  	//Mark Value
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_MARK : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_CYCLE_REG: 	  	//Data write cycle
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_CYCLE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_MODE_REG: 	  	//ADD mode
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_MODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_NUM_REG: 	  	//Amount of non-transferred data
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_NUM : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_MASK_REG: 	  	//Write mask pattern
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_MASK : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_CODE_REG: 	  	//Last processed VIFcode
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_CODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_ITOPS_REG: 	  	//Next ITOP value
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_ITOPS : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
/*
		case VIF0_BASE_REG: 	  	//Base address of double buffer
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_BASE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_OFST_REG: 	  	//Offset of double buffer
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_OFST : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_TOPS_REG: 	  	//Next TOP value/data write address
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_TOPS : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_ITOP_REG: 	  	//ITOP value
			DMAC_LOG("%8.8lx:[VIF0][WR]VIF0_ITOP : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_TOP_REG: 	  	//TOP value
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_TOP : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_R0: 	  	//Filling data R0 (Row register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_R0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_R1: 	  	//Filling data R1 (Row register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_R1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_R2: 	  	//Filling data R2 (Row register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_R2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_R3: 	  	//Filling data R3 (Row register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_R3 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_C0: 	  	//Filling data C0 (Col register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_C0 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_C1: 	  	//Filling data C1 (Col register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_C1 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_C2: 	  	//Filling data C2 (Col register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_C2 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case VIF0_C3: 	  	//Filling data C3 (Col register)
			VIF_LOG("%8.8lx:[VIF0][WR]VIF0_C3 : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
*/

		default: 

			EMU_LOG("%8.8lx:[VIF0][WR]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);

		break;
#endif
	

	}
}


/* Read from VIF0 Registers */
void	VIF0_Read ( _u64 fPhysicalAddress , _u128 * fVal , _u8 fDataType ) {

	_u32	boucle;
	_u64	RealAdd;


#ifdef VIF_LOG
	switch ( fPhysicalAddress  ) {

		case VIF0_STAT_REG: 	  //Status
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_STAT : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->stat);
		break;

		case VIF0_FBRST_REG: 	  	//Operation Control
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_FBRST : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->fbrst);
		break;

		case VIF0_ERR_REG: 	  	//Error status
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_ERR : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->err);
		break;

		case VIF0_MARK_REG: 	  	//Mark Value
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_MARK : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->mark);
		break;

		case VIF0_CYCLE_REG: 	  	//Data write cycle
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_CYCLE : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->cycle);
		break;

		case VIF0_MODE_REG: 	  	//ADD mode
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_MODE : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->mode);
		break;

		case VIF0_NUM_REG: 	  	//Amount of non-transferred data
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_NUM : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->num);
		break;

		case VIF0_MASK_REG: 	  	//Write mask pattern
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_MASK : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->mask);
		break;

		case VIF0_CODE_REG: 	  	//Last processed VIFcode
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_CODE : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->code);
		break;

		case VIF0_ITOPS_REG: 	  	//Next ITOP value
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_ITOPS : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->itops);
		break;
/*
		case VIF0_BASE_REG: 	  	//Base address of double buffer
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_BASE : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->base);
		break;

		case VIF0_OFST_REG: 	  	//Offset of double buffer
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_OFST : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->ofst);
		break;

		case VIF0_TOPS_REG: 	  	//Next TOP value/data write address
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_TOPS : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->tops);
		break;

		case VIF0_ITOP_REG: 	  	//ITOP value
			DMAC_LOG("%8.8lx:[VIF0][RD]VIF0_ITOP : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->itop);
		break;

		case VIF0_TOP_REG: 	  	//TOP value
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_TOP : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->top);
		break;

		case VIF0_R0: 	  	//Filling data R0 (Row register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_R0 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->r0);
		break;

		case VIF0_R1: 	  	//Filling data R1 (Row register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_R1 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->r1);
		break;

		case VIF0_R2: 	  	//Filling data R2 (Row register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_R2 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->r2);
		break;

		case VIF0_R3: 	  	//Filling data R3 (Row register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_R3 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->r3);
		break;

		case VIF0_C0: 	  	//Filling data C0 (Col register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_C0 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->c0);
		break;

		case VIF0_C1: 	  	//Filling data C1 (Col register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_C1 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->c1);
		break;

		case VIF0_C2: 	  	//Filling data C2 (Col register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_C2 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->c2);
		break;

		case VIF0_C3: 	  	//Filling data C3 (Col register)
			VIF_LOG("%8.8lx:[VIF0][RD]VIF0_C3 : 0x%8.8lx ", (long)MipsCpu.PC,VIF[ VIF0 ].Reg->c3);
		break;
*/

		default: {
			EMU_LOG("%8.8lx:[VIF0][RD]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
		break;
		}

	}
#endif

	RealAdd 		= fPhysicalAddress & 0xFFFF;

	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryReadDQ 	( SpecialAreaMemory, fVal, RealAdd );break;
					
		case DATA_DW:	MipsMemoryReadDW	( SpecialAreaMemory, fVal, RealAdd );break;

	    case DATA_HW:	MipsMemoryReadHW	( SpecialAreaMemory, fVal, RealAdd );break;
		
		case DATA_W:	MipsMemoryReadW 	( SpecialAreaMemory, fVal, RealAdd );break;

		case DATA_B:	MipsMemoryReadB		( SpecialAreaMemory, fVal, RealAdd );break;

		default:		EMU_LOG("%8.8lx:[VIF0][RD][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
						MipsCpu.halt = 1;break;
	}






}












