/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>


#include <string.h>

#include "Common.h"

#include "Loader_elf.h"


elf_header ico_elf_hdr;
elf_program_header * ico_elf_prg_hdr ;
elf_section_header * ico_elf_sec_hdr ;
long ElfFileSize;

unsigned long	ElfStartAdress;



u8 ElfLoadHeader( char * fbuff) {
	

	u8 Result;

		
	Result = 0;

	
	memcpy( &ico_elf_hdr, fbuff, sizeof(struct  _elf_header ) );

	ElfStartAdress = ico_elf_hdr.entry_point_add;




	if((ico_elf_hdr.num_of_prg_hdr * ico_elf_hdr.prg_hdr_entry_size) != 0) {

		ico_elf_prg_hdr = (struct _elf_program_header *) malloc(ico_elf_hdr.num_of_prg_hdr * ico_elf_hdr.prg_hdr_entry_size);

		memcpy( (u8 *)ico_elf_prg_hdr, (u8*)(fbuff) + ico_elf_hdr.start_prg_hdr, ico_elf_hdr.num_of_prg_hdr * ico_elf_hdr.prg_hdr_entry_size );

	}
	else
		ico_elf_prg_hdr = NULL;

	



	if((ico_elf_hdr.num_of_sec_hdr * ico_elf_hdr.sec_hdr_entry_size) != 0) {

		ico_elf_sec_hdr = (struct _elf_section_header *) malloc(ico_elf_hdr.num_of_sec_hdr * ico_elf_hdr.sec_hdr_entry_size);

		memcpy( (u8 *)ico_elf_sec_hdr, (u8*)(fbuff) + ico_elf_hdr.start_sct_hdr, ico_elf_hdr.num_of_sec_hdr * ico_elf_hdr.sec_hdr_entry_size);
		
	}
	else
		ico_elf_sec_hdr = NULL;


	

	SysPrintf("[LOADER][ELF][HEADER]Read");

	

	return Result;
}

void ElfFreeHeader( void ) {

	if ( ico_elf_prg_hdr != NULL ) free ( ico_elf_prg_hdr ) ;
	if ( ico_elf_sec_hdr != NULL ) free ( ico_elf_sec_hdr );

}


void CopyProgramToMemory( char * fbuff) {
	
	u32 Addr;

	u64 Index2;
	u64 ReadError;





	u32	FirstAdress = 0xFFFFFFFF;
	u32	EndAdress = 0x0;

	u16	NbPrgSectionLoaded = 0;
	

	
	ReadError = 0;


	if ( ico_elf_prg_hdr != NULL ) {
	
		for ( Index2 = 0 ; Index2 <  ico_elf_hdr.num_of_prg_hdr ; Index2++ ) {

			switch ( ico_elf_prg_hdr[Index2].segment_type ) {

				case 0x1:
			
					


					Addr = ico_elf_prg_hdr[Index2].physic_add_of_segment;

					memcpy( (u8*)PSM(Addr), (u8*)(fbuff) + ico_elf_prg_hdr[Index2].offs_from_start_to_prg_seg, ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima ); 

#ifdef LOADER_LOG
					SysPrintf("[LOADER][ELF]Prog %d : Loaded (size:%8.8lx)",Index2,ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima);

#endif			
					NbPrgSectionLoaded++;

				break;


				default:

					SysPrintf("[LOADER][ELF]Prog %d : Unknown (size:%8.8lx)",Index2,ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima);

				break;
	
			}
		
		}
	
	}

	if (ReadError != 0 )	{

		SysPrintf("[LOADER][ELF][HEADER]Prog : Done with %lld errors",ReadError);


	}
	else	{

	SysPrintf("[LOADER][ELF][HEADER]Prog : Done ( %d sections Loaded )",NbPrgSectionLoaded);

	}
	


}


void CopySectionToMemory ( char * fbuff) {

	u32 Addr = 0;
	u32 index;

	u64 ReadError;	

	u16	NbPrgSectionLoaded = 0;	

	ReadError = 0;

	if ( ico_elf_prg_hdr != NULL ) {

/*
		fseek( ElfFile,  ico_elf_hdr.start_sct_hdr, SEEK_SET );
*/		
		for ( index = 0 ; index <  ico_elf_hdr.num_of_sec_hdr ; index++ ) {

			

			switch ( ico_elf_sec_hdr[index].type ) {

			case SHT_NULL:

#ifdef LOADER_LOG
				SysPrintf("[LOADER][ELF]Section %d : NULL Loaded ( %d )",index,ico_elf_sec_hdr[index].size);
#endif


			break;


			case SHT_PROGBITS:
				Addr =  ico_elf_sec_hdr[index].sect_start_add & 0x1FFFFFFF;

				memcpy( (u8*)PSM(Addr), (u8*)(fbuff) + ico_elf_sec_hdr[index].offset_from_start, ico_elf_sec_hdr[index].size ); 


				/*
				fseek( ElfFile,   ico_elf_sec_hdr[index].offset_from_start  , SEEK_SET );	


				readOK = fread((u8 *)(MipsMemory + Addr), 1, ico_elf_sec_hdr[index].size, ElfFile);
*/			
				NbPrgSectionLoaded++;	

#ifdef LOADER_LOG
				SysPrintf("[LOADER][ELF]Section %d : PROGBITS Loaded ( %d )",index,ico_elf_sec_hdr[index].size);
#endif

 			break;


			case SHT_NOBITS:
#ifdef LOADER_LOG
				SysPrintf("[LOADER][ELF]Section %d : NOBITS Loaded ( %d )",index,ico_elf_sec_hdr[index].size);
#endif
			break;
			


			default:

				SysPrintf("[LOADER][ELF]Section %d : Unknown  %8.8lx (size:%8.8lx)",index,ico_elf_sec_hdr[index].type,ico_elf_sec_hdr[index].size);


			break;
			
	
			}
		
		}

	}
						

	SysPrintf("[LOADER][ELF][HEADER]Section : Done ( %d sections Loaded )",NbPrgSectionLoaded);



}



