/*
 *  Original code from sjdata.irx by Sjeep (C) 2002
 *  Modified by Florin for PCSX2 emu
 */

#include "Common.h"

#include "SJdata.h"
#include "CDVDisodrv.h"
#include "RPC/Rpc_services.h"

#define FILE_BUFF_LEN	0x4000

struct datafile_header {
	char identity[4];
	int num_entries;
	int pad[2];
};

struct datafile_index_entry {
	char filename[56];
	int file_length;
	char* file_offset;
};

struct fdtable {
	int length;
	int offset;
	int filepos;
};

static struct fdtable fd_table[16];
static int fd_used[16];
static int files_open = 0;
static struct datafile_header* header;
static struct datafile_index_entry* entry;
static char* header_buffer;
static char* file_buffer;

static int datafile_fd=-1;
char datafile_name[256];

/////////////////////////////////////////////////////////
//
//   SjDATA decode routines
//
/////////////////////////////////////////////////////////

int decode(char* in, char* out, int length, char acc, int beginning){
	int i = 0;

	if(beginning) {
		out[0] = in[0];
		acc = in[0];
		i++;
	}

	for(; i < length; i++) {
		out[i] = in[i] ^ acc;
		acc = in[i];
	}

	return 0;
}

void unscramble_ascii(char* string){
	int i;
	int len = strlen(string)+1;

	for(i=0;i<len;i++)
		string[i] -= 43;
}

static int inited=FALSE;

///////////////////////////////////////////////////////////
//	SjDATA_init
///////////////////////////////////////////////////////////
void SjDATA_init(){
	char* temp_buffer;
	char temp_string[100];
	int header_size;
	int i;

#ifdef RPC_LOG
	RPC_LOG("[SjDATA:init] SjDATA - by Nick Van Veen (aka Sjeep)\n");
	RPC_LOG("[SjDATA:    ] Initializing sjdata file driver.\n");
#endif

 	file_buffer = AllocSysMemory(0, FILE_BUFF_LEN+1, NULL); // allocate a 16kb buffer
  	if(file_buffer == NULL) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Failed to allocate memory! (file buffer)\n");
#endif
		return;
	}

	memset(fd_table, 0, sizeof(struct fdtable) * 16);
	memset(fd_used, 0, 16*4);
	memset(PSXM((int)file_buffer), 0, FILE_BUFF_LEN+1);

	// open datafile, read in header
	temp_buffer = AllocSysMemory(0,64,0);
	if(temp_buffer == NULL) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Failed to allocate memory! (temporary)\n");
#endif
		return;
	}
	datafile_fd = CDVDFS_open(datafile_name, 1);//O_RDONLY
	if(datafile_fd < 0) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Failed to open datafile: %s\n",datafile_name);
#endif
		return;
	}
	if(CDVDFS_read(datafile_fd,PSXM((int)temp_buffer),64) != 64) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Failed to read datafile header!\n");
#endif
		return;
	}

	header = (struct datafile_header*)PSXM((int)temp_buffer);

	// make sure file is a sjdata datafile
	strncpy((char *)temp_string,(const char *)header->identity,4);
	temp_string[4] = '\0';
	if(!(strstr(temp_string,"SJDF"))) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Invalid datafile!\n");
#endif
		return;
	}

	header_size = sizeof(struct datafile_header)+(header->num_entries*sizeof(struct datafile_index_entry));
	FreeSysMemory(temp_buffer);
	header_buffer = AllocSysMemory(0, header_size, NULL);
	if(header_buffer == NULL) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Failed to allocate memory! (header buffer)\n");
#endif
		return;
	}

	// read in header
	CDVDFS_lseek(datafile_fd, 0, SEEK_SET);
	if(CDVDFS_read(datafile_fd, PSXM((int)header_buffer), header_size) != header_size) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] Failed to read datafile header!\n");
#endif
		return;
	}

	header = (struct datafile_header *)PSXM((int)header_buffer);
	entry = (struct datafile_index_entry *)((char*)PSXM((int)header_buffer) + sizeof(struct datafile_header));

	for(i=0; i < header->num_entries; i++) unscramble_ascii(entry[i].filename);

	CDVDFS_close(datafile_fd);
	datafile_fd = -1;

#ifdef RPC_LOG
	RPC_LOG("[SjDATA:    ] Initilisation complete. Datafile: %s\n", datafile_name);
#endif

	inited=TRUE;

	return;
}

///////////////////////////////////////////////////////////
//	SjDATA_lseek
///////////////////////////////////////////////////////////
int SjDATA_lseek( int fd, int offset, int whence ){
	int i=fd;

	if (!inited)	SjDATA_init();

#ifdef RPC_LOG
	RPC_LOG("[SjDATA:seek] fd=%d, offset=%d, whence=%s\n",
		i, offset, whence==SEEK_SET?"SEEK_SET":whence==SEEK_CUR?"SEEK_CUR":"SEEK_END");
#endif

	if(i >= 16) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:lseek] ERROR: File does not appear to be open!\n");
#endif
		return -1;
	}

	switch(whence) {
		case SEEK_SET:
			fd_table[i].filepos = offset;
			return fd_table[i].filepos;
		case SEEK_CUR:
			fd_table[i].filepos += offset;
			return fd_table[i].filepos;
		case SEEK_END:
			fd_table[i].filepos = fd_table[i].length + offset;
			return fd_table[i].filepos;
	}

	return -1;
}

///////////////////////////////////////////////////////////
//	SjDATA_open
///////////////////////////////////////////////////////////
int SjDATA_open(char *name, int mode){
	int i,j;

	if (!inited)	SjDATA_init();

#ifdef RPC_LOG
	RPC_LOG("[SjDATA:open] name=\"%s\", mode=%d\n", name, mode );
#endif

	// check to see if filename exists in datafile
	for(i=0; i < header->num_entries; i++)
		if(strcmp(entry[i].filename, name)==0)
			break;

	if(i >= header->num_entries) return -1;

	if(mode != 1) return -2;	//O_RDONLY

	// set up a new file descriptor
	for(j=0; j < 16; j++) if(fd_used[j] == 0) break;
	if(j >= 16) return -3;

	fd_used[j] = 1;
	files_open++;
#ifdef RPC_LOG
	RPC_LOG("[SjDATA:    ] internal fd=%d\n", j);
#endif

	fd_table[j].length = entry[i].file_length;
	fd_table[j].offset = (int)entry[i].file_offset;
	fd_table[j].filepos = 0;

   	return j;
}

///////////////////////////////////////////////////////////
//	SjDATA_read
///////////////////////////////////////////////////////////
int SjDATA_read( int fd, char *buffer, int size ){
	int i=fd;
	int bytes_read;
	int file_position;

	if (!inited)	SjDATA_init();

#ifdef RPC_LOG
	RPC_LOG("[SjDATA:read] fd=%d, buffer=0x%08X, size=%d\n",
		i, buffer, size);
#endif

   	// open main datafile if not already open
   	if(datafile_fd == -1) {
		datafile_fd = CDVDFS_open(datafile_name, 1);//SCE_RDONLY
		if(datafile_fd < 0){
#ifdef RPC_LOG
			RPC_LOG("[SjDATA:    ] Failed to open datafile: %s\n", datafile_name);
#endif
			return -1;
		}
	}
	
	if(i >= 16) {
#ifdef RPC_LOG
		RPC_LOG("[SjDATA:    ] ERROR: File does not appear to be open!\n");
#endif
		return -1;
	}

	if(size > (fd_table[i].length-fd_table[i].filepos))
		size = fd_table[i].length - fd_table[i].filepos;

	// read data into temp buffer and decode into output buffer
	if(size > FILE_BUFF_LEN) {

		int lenread = FILE_BUFF_LEN;
		char *tempptr;
		int tempret;

		bytes_read = 0;
		tempptr = buffer;
		while(lenread) {

			if(fd_table[i].filepos)
				file_position = fd_table[i].offset + fd_table[i].filepos - 1;
			else
				file_position = fd_table[i].offset;
			
			if((tempret = CDVDFS_lseek(datafile_fd, file_position, SEEK_SET)) != file_position ) {
#ifdef RPC_LOG
			RPC_LOG("[SjDATA:    ] ERROR: Read lseek failed! Pos = 0x%X Offset = 0x%X Ret = %d\n",file_position, fd_table[i].offset, tempret);
#endif
				return -2;
			}

//			printf(" ---- tempptr = 0x%X, bytes_read = %d, lenread = %d, file_position = 0x%X\n",tempptr,bytes_read,lenread,file_position);

			if(CDVDFS_read(datafile_fd, PSXM((int)file_buffer), lenread + 1) != (lenread + 1)) {
#ifdef RPC_LOG
			RPC_LOG("[SjDATA:    ] ERROR: Failed to read file! (packet mode)\n");
#endif
				return -3;
			}

			if(fd_table[i].filepos)
				decode(PSXM((int)file_buffer+1),tempptr,lenread,*(char*)PSXM((int)file_buffer),0);
			else
				decode(PSXM((int)file_buffer),tempptr,lenread,*(char*)PSXM((int)file_buffer),1);

			tempptr += lenread;
			bytes_read += lenread;
			fd_table[i].filepos += lenread;

			if((size - bytes_read) < FILE_BUFF_LEN)
				lenread = size - bytes_read;

		}

	} else {
	
		// if not at start of file, must load byte previous to 'filepos' to be used for decoding
		if(fd_table[i].filepos) {
			file_position = fd_table[i].offset + fd_table[i].filepos - 1;
			size++;
		} else
			file_position = fd_table[i].offset;

		if(CDVDFS_lseek(datafile_fd, file_position, SEEK_SET) != file_position) {
#ifdef RPC_LOG
   			RPC_LOG("[SjDATA:    ] ERROR: Read lseek failed! Pos = 0x%X\n",file_position);
#endif
			return -2;
		}

//		printf(" ---- buffer = 0x%X, bytes_read = %d, file_position = 0x%X\n",buffer,bytes_read,file_position);
		CDVDFS_read(datafile_fd, PSXM((int)file_buffer), size);
		
		if(fd_table[i].filepos) size--;
		bytes_read = size;

		if(fd_table[i].filepos)
			decode(PSXM((int)file_buffer+1), buffer, bytes_read,*(char*)PSXM((int)file_buffer),0);
		else
			decode(PSXM((int)file_buffer), buffer, bytes_read,*(char*)PSXM((int)file_buffer),1);

		fd_table[i].filepos += bytes_read;

	}

   	return bytes_read;
}

///////////////////////////////////////////////////////////
//	SjDATA_write
///////////////////////////////////////////////////////////
int SjDATA_write( int fd, char *buffer, int size){

	if (!inited)	SjDATA_init();

	if(size == 0) return 0;
	else 		 return -1;
}

///////////////////////////////////////////////////////////
//	SjDATA_close
///////////////////////////////////////////////////////////
int SjDATA_close( int kernel_fd ){
	int i=kernel_fd;

	if (!inited)	SjDATA_init();

#ifdef RPC_LOG
	RPC_LOG("[SjDATA:close] fd=%d\n", kernel_fd );

	if (fd_used[i]==0){
		RPC_LOG("[SjDATA:      ] ERROR: File does not appear to be open!\n");
	}
#endif

	fd_used[i] = 0;
	files_open--;

	if(!files_open) {
		CDVDFS_close(datafile_fd);
		datafile_fd = -1;
	}

   return 0;
}
