/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>


#include "nSX2.h"

#include "r5900_main.h"
 
#include "bios_syscalls.h"
#include "bios_syscalls_table.h"
 
#include "PS2Edefs.h"

#include "Sif.h"

/* I must declare a function to avoid this kind of include */
#include "DmaC.h"

/* must create function to avoid this kind of declaration */
#include "IntrC.h"


#include "GS.h"






/**********************************************************************************************************
									DMA CHANNEL CONTROL
***********************************************************************************************************/


/*
_EnableDmac	

Description :
This system call enables DMA termination interrupts for the specified channel.

Params:
a0 : Channel Number

Return:
v0 : 	0 Already enabled
		1 Enabled
*/
void SYSCALL__EnableDmac( )  					{
	char temp_string[256];

    _u32  newValue;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]_EnableDmac : channel n %ld ",(long)MipsCpu.PC,
		(long)MipsCpu.GPR[CPU_REG_A0].lo );
#endif

	
	/* compute the new value */
    newValue = ( 0x10000 << MipsCpu.GPR[CPU_REG_A0].lo ) &  DmaC.Reg->stat;
    
    /*If the  is already Enabled */
    if ( newValue != 0 ) {
		/* return 0 */
		MipsCpu.GPR[CPU_REG_V0].lo =  0;
     }
    /* Else enable the Interrupt for the specified channel */
    else {
        /* Write the new value of DMA_STAT */
        DmaC.Reg->stat = DmaC.Reg->stat | ( 0x10000 << MipsCpu.GPR[CPU_REG_A0].lo ) ;

		/* return 1 */
        MipsCpu.GPR[CPU_REG_V0].lo =  1;
    }
 		
	INCREMENT_PC();

} 


/*
_DisableDmac	ENTRY_23	80000980

Description :
This system call disables DMA termination interrupts for the specified channel.

Params:
a0 : Channel Number

Return:
v0 : 	0 Already disabled
		1 disabled

*/
void SYSCALL__DisableDmac( ) 					{ 

	char temp_string[256];

    _u32  newValue;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]_DisableDmac : channel n %ld ",(long)MipsCpu.PC,
		(long)MipsCpu.GPR[CPU_REG_A0].lo );
#endif	
	
	/* compute the new value */
    newValue = ( 0x10000 << MipsCpu.GPR[CPU_REG_A0].lo ) &  (DmaC.Reg->stat);
   
    /*If the Interrupt is already Enabled */
    if ( newValue == 0 ) {
		/* return 0 */
		MipsCpu.GPR[CPU_REG_V0].lo =  0;
     }
    /* Else enable the Interrupt for the specified channel */
    else {
        /* Write the new value of DMA_STAT */

		DmaC.Reg->stat = DmaC.Reg->stat & ~( 0x10000 << MipsCpu.GPR[CPU_REG_A0].lo ) ;

		/* return 1 */
        MipsCpu.GPR[CPU_REG_V0].lo =  1;
    }
 		
	INCREMENT_PC();





}


void SYSCALL__iDisableDmac( ) { 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]_iDisableDmac : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
} 


void SYSCALL__iEnableDmac( ) { 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]_iEnableDmac : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}     



/**********************************************************************************************************
				DMA CHANNEL HANDLER									
***********************************************************************************************************/


void SYSCALL_EnableDmacHandler( ) 				{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]EnableDmacHandler : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}   



void SYSCALL_DisableDmacHandler( ) 				{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]DisableDmacHandler : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}   



/*
AddDmacHandler	ENTRY_18	80001c78

Description :
Adds a DMA interrupt handler by DMA channel. handler is the address of the handler 
that will be called at the time of the DMA interrupt. channel is the DMA channel 
number and will be passed to the handler in argument ch. next is the address of 
a previously registered DMA interrupt handler. If next is 0, handler will
be registered as the first interrupt handler. If next is -1, handler will be 
registered as the last interrupt handler.

Note: When a DMA interrupt is generated, the GPRs are saved before the interrupt 
handler is called. When the handler has finished executing, the GPRs are 
restored. However, note that there is no save/restore of FPU registers.

When writing interrupt handlers, note that interrupts are disabled and that 
different APIs can be used. Refer to \overview\eekernel for information on 
these issues.

If an interrupt handler returns -1, subsequently registered interrupt handlers 
will not be called.


Params :
a0: channel number
a1: Address handler
a2: position of the handler ( not used yet )

Return:
normal termination : Handler ID
Error : -1
*/
void SYSCALL_AddDmacHandler( ) {

	char temp_string[256];
	_u8		DMA_channel;
	_u64	DMA_handler_addr;
	_u64	Dma_next_handler;

	DMA_channel 		= MipsCpu.GPR[CPU_REG_A0].lo & 0xFF;
	DMA_handler_addr	= MipsCpu.GPR[CPU_REG_A1].lo;
	Dma_next_handler	= MipsCpu.GPR[CPU_REG_A2].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]AddDmacHandler channel n %ld , Addr : 0x%8.8lx , next : 0x%8.8lx",(long)MipsCpu.PC,
		 DMA_channel, DMA_handler_addr, Dma_next_handler);
#endif	

	MipsCpu.GPR[CPU_REG_V0].lo =  BIOS_AddDmacHandler ( DMA_channel, DMA_handler_addr,Dma_next_handler ) ;
	
	INCREMENT_PC();
} 


/*
RemoveDmacHandler	ENTRY_19	80001cb8
*/
void SYSCALL_RemoveDmacHandler( ) 				{ 

	char temp_string[256];
	_u8		DMA_channel;
	_u64	DMA_handler_addr;
	_u64	Dma_next_handler;

	DMA_channel 		= MipsCpu.GPR[CPU_REG_A0].lo & 0xFF;
	DMA_handler_addr	= MipsCpu.GPR[CPU_REG_A1].lo;
	Dma_next_handler	= MipsCpu.GPR[CPU_REG_A2].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RemoveDmacHandler channel n %ld , hID : %ld",(long)MipsCpu.PC,
		 DMA_channel, DMA_handler_addr );
#endif	

	MipsCpu.GPR[CPU_REG_V0].lo = BIOS_RemoveDmacHandler(DMA_channel, DMA_handler_addr );
	
	INCREMENT_PC();

}   


void SYSCALL_iDisableDmacHandler( ) 			{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iDisableDmacHandler : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}     


void SYSCALL_iEnableDmacHandler( ) 				{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iEnableDmacHandler : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}     


/******************************************************************************/


_u32 BIOS_AddDmacHandler ( _u32 fChannel, _u32 fHandlerAdd, _u32 fNext ) {

	_u32	ReturnValue;

	LinkedAddress * FirstHandlerAdd;
	LinkedAddress * CurrentHandlerAdd;
	LinkedAddress * SavedHandlerAdd;

	FirstHandlerAdd 	= IntC.IntByDMA[ fChannel ].HandlerAdd;

	/* How to Register... */
	switch ( (_u32)fNext ) {
		/* Registered as first handler */
		case 0 :
	
			SavedHandlerAdd = FirstHandlerAdd;				
			CurrentHandlerAdd = ( LinkedAddress * ) malloc ( sizeof ( LinkedAddress ) );

			IntC.IntByDMA[ fChannel ].Status 	= 1 ;

			CurrentHandlerAdd->ID	= IntC.IntByDMA[ fChannel ].Numero_ID;
			CurrentHandlerAdd->Add	= (_u32) fHandlerAdd;
			CurrentHandlerAdd->next = SavedHandlerAdd;

			IntC.IntByDMA[ fChannel ].HandlerAdd 	= CurrentHandlerAdd;
			IntC.IntByDMA[ fChannel ].Numero_ID		= IntC.IntByDMA[ fChannel ].Numero_ID + 1;
			IntC.IntByDMA[ fChannel ].Nb_handler	= IntC.IntByDMA[ fChannel ].Nb_handler + 1;

			ReturnValue = CurrentHandlerAdd->ID;

		break;

		/* Registred as Last handler */
		case (_u32)0xFFFFFFFF :
			SavedHandlerAdd = FirstHandlerAdd;

			IntC.IntByDMA[ fChannel ].Status 	= 1 ;
			
			while ( SavedHandlerAdd->next != NULL )
				SavedHandlerAdd = SavedHandlerAdd->next;

			SavedHandlerAdd->next = ( LinkedAddress * ) malloc ( sizeof ( LinkedAddress ) );
			
			CurrentHandlerAdd = SavedHandlerAdd->next;

			CurrentHandlerAdd->Add	= (_u32) fHandlerAdd;
			CurrentHandlerAdd->ID	= IntC.IntByDMA[ fChannel ].Numero_ID;
			CurrentHandlerAdd->next = NULL;
				
			IntC.IntByDMA[ fChannel ].Numero_ID		= IntC.IntByDMA[ fChannel ].Numero_ID + 1;
			IntC.IntByDMA[ fChannel ].Nb_handler	= IntC.IntByDMA[ fChannel ].Nb_handler + 1;

			ReturnValue = CurrentHandlerAdd->ID;
	
		break;

		/* Register after next */
		default:

#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]AddDmacHandler : UnHandle Case Next",(long)MipsCpu.PC );
#endif	
			MipsCpu.halt = 1;
	
			ReturnValue = (_u32)0xFFFFFFFF;
		break;

	}

	return ReturnValue;
}


_u32	BIOS_RemoveDmacHandler(_u32 fChannel, _u32 fHandlerID ) {

	_u8		FoundHandler = FALSE;
	_u32	ReturnValue;


	LinkedAddress * PreviousHandlerAdd;	
	LinkedAddress * CurrentHandlerAdd;
	LinkedAddress * NextHandlerAdd;	

	PreviousHandlerAdd 	= IntC.IntByDMA[ fChannel ].HandlerAdd;
	CurrentHandlerAdd 	= IntC.IntByDMA[ fChannel ].HandlerAdd;
	if ( CurrentHandlerAdd != NULL ) NextHandlerAdd		= CurrentHandlerAdd->next;


	while ( CurrentHandlerAdd != NULL ) {


		if ( CurrentHandlerAdd->ID == fHandlerID ) {

			if ( PreviousHandlerAdd != NULL ) PreviousHandlerAdd->next = NextHandlerAdd;
			
			free ( CurrentHandlerAdd );
			CurrentHandlerAdd = NULL;

			IntC.IntByDMA[ fChannel ].Nb_handler = IntC.IntByDMA[ fChannel ].Nb_handler - 1;

			if ( IntC.IntByDMA[ fChannel ].Nb_handler == 0 ) IntC.IntByDMA[ fChannel ].HandlerAdd = NULL;

			FoundHandler = TRUE;
		}
		else {

			PreviousHandlerAdd = CurrentHandlerAdd;
			if ( CurrentHandlerAdd != NULL ) CurrentHandlerAdd = CurrentHandlerAdd->next;
			if ( CurrentHandlerAdd != NULL ) NextHandlerAdd = CurrentHandlerAdd->next;

		}


	}




	if ( FoundHandler == TRUE )  
		ReturnValue	= IntC.IntByDMA[ fChannel ].Nb_handler;
	else 
		ReturnValue	= (_u32)0xFFFFFFFF;
	

	return ReturnValue;
	
}


