/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>


#include "nSX2.h"

#include "r5900_main.h"
 
#include "bios_syscalls.h"
#include "bios_syscalls_table.h"
 
#include "PS2Edefs.h"

#include "Sif.h"

/* I must declare a function to avoid this kind of include */
#include "DmaC.h"

/* must create function to avoid this kind of declaration */
#include "IntrC.h"


#include "GS.h"


#include "LL_Threads/LinkedList.h"
/**********************************************************************************************************
									THREADS
***********************************************************************************************************/

/*
RFU060_InitializeMainThread

Status : OK
*/
void	SYSCALL_RFU060() {
	_u32 GPreg, Stack, StackSize, Args, Root;
	PS2_Thread	* InitPS2Thread;
	LLThread	* TempBiosThread;

	GPreg		= (_u32)MipsCpu.GPR[CPU_REG_A0].lo;	
	Stack		= (_u32)MipsCpu.GPR[CPU_REG_A1].lo;
	StackSize	= (_u32)MipsCpu.GPR[CPU_REG_A2].lo;
	Args		= (_u32)MipsCpu.GPR[CPU_REG_A3].lo;
	Root		= (_u32)MipsCpu.GPR[CPU_REG_T0].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU060_InitializeMainThread ",(long)MipsCpu.PC );
#endif	
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]InitializeMainThread : GP:0x%8.8lx Stack:0x%8.8lx",Ps2Bios.PS2_Sched.RunningIDThread,
			GPreg,Stack);
	THREAD_LOG("(%d)[BIOS]|StackSize:0x%8.8lx Args:0x%8.8lx Root:0x%8.8lx",Ps2Bios.PS2_Sched.RunningIDThread,
			StackSize,Args,Root);
#endif
	



	/* Ok... let's rock baby */
	/*  \/  new Code */

	if ( (_s32)StackSize < 0 ) {
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]|Fixed StackSize:0x%8.8lx ",Ps2Bios.PS2_Sched.RunningIDThread,StackSize);
#endif
		StackSize = 4096;
	}

	if ( (_u32)Stack == (_u32)0xFFFFFFFF ) {
		Stack = (_u32)0x02000000 - StackSize - 4096;
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]|Fixed StackAdd:0x%8.8lx ",Ps2Bios.PS2_Sched.RunningIDThread,Stack);
#endif
	}





	/* destroy the current Thread */
	Ps2Bios.PS2_Sched.Running_Thread->thread->status = THS_DESTROY;

	/* Create a new thread */
	InitPS2Thread = ( PS2_Thread * ) malloc ( sizeof( PS2_Thread ) ) ;

	/* Put in READY State, it will be put in Run state by the Scheduler */
	InitPS2Thread->status  			= THS_READY;
	InitPS2Thread->entry_add  		= MipsCpu.PC + 4;
	InitPS2Thread->initPriority		= 0;			//Maximum Priority
	InitPS2Thread->currentPriority	= 0;			//
	InitPS2Thread->stackSize		= (_u32)StackSize;
	InitPS2Thread->stack_add  		= (_u32)Stack;

	
	TempBiosThread = THREADS_Create( InitPS2Thread,	InitPS2Thread->status );

	/* Save the GPR State of the Thread */
	memcpy( TempBiosThread->GPR, MipsCpu.GPR, sizeof( _u128) * 34 );
	memcpy( TempBiosThread->FPR1, MipsCpu.FPR1, sizeof( _u32) * 32 );

	//the Main Thread Param
	TempBiosThread->GPR[CPU_REG_SP].lo = (_u32)Stack + StackSize;
	TempBiosThread->GPR[CPU_REG_S8].lo = (_u32)Stack + StackSize;
	TempBiosThread->GPR[CPU_REG_GP].lo = (_u32)GPreg;

	//it's the return Value of the Syscall
	TempBiosThread->GPR[CPU_REG_V0].lo = (_u32)Stack + StackSize;	
	//TempBiosThread->GPR[CPU_REG_V0].lo = (_u32)0x81F00000;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]|ReturnValue:0x%8.8lx",(long)MipsCpu.PC,TempBiosThread->GPR[CPU_REG_V0].lo );
#endif

	//the execution go on after the syscall
	TempBiosThread->PC = MipsCpu.PC + 4;
	
	/* Original Code */
	/* MipsCpu.GPR[CPU_REG_V0].lo = 0x81F00000; */

	INCREMENT_PC();

	// Update the Scheduler
	UpdatePS2Scheduler();


	//MipsCpu.halt = 1; 

}

/*
RFU061_InitializeHeapArea

Status : OK
*/
void	SYSCALL_RFU061() {
	_u32	End, HeapSize;


	End			= (_u32)MipsCpu.GPR[CPU_REG_A0].lo;	
	HeapSize	= (_u32)MipsCpu.GPR[CPU_REG_A1].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU061_InitializeHeapArea",(long)MipsCpu.PC );
#endif	
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]InitializeHeapArea : EndAdd:0x%8.8lx HeapSize:0x%8.8lx",Ps2Bios.PS2_Sched.RunningIDThread,
			End,HeapSize);
#endif
	
/* NE PAS EFFACER
	if ( (_u32)HeapSize == (_u32)0xFFFFFFFF ) {
		Ps2Bios.PS2_Sched.Running_Thread->EndOfHeap = 
			Ps2Bios.PS2_Sched.Running_Thread->thread->stack_add ;
	} else {
	 	Ps2Bios.PS2_Sched.Running_Thread->EndOfHeap =	End + HeapSize ;
	}  	
*/

	if ( (_u32)HeapSize == (_u32)0xFFFFFFFF )
		HeapSize = Ps2Bios.PS2_Sched.Running_Thread->thread->stack_add - End;


	Ps2Bios.PS2_Sched.Running_Thread->HeapSize 	= HeapSize;
	Ps2Bios.PS2_Sched.Running_Thread->EndOfHeap = End + HeapSize;

	
	MipsCpu.GPR[CPU_REG_V0].lo = (_u32)Ps2Bios.PS2_Sched.Running_Thread->EndOfHeap;
	INCREMENT_PC();

}



void SYSCALL_EndOfHeap( ) 						{ 


#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]EndOfHeap",(long)MipsCpu.PC );
#endif	
	
 //	MipsCpu.GPR[CPU_REG_V0].lo = 0x80000000;//Original Value
 	
	MipsCpu.GPR[CPU_REG_V0].lo = (_u32)Ps2Bios.PS2_Sched.Running_Thread->EndOfHeap;

 	INCREMENT_PC();
}






void SYSCALL_CreateThread( ) 					{ 
	_u32			ThreadAddr;
	PS2_Thread 		* myTh = NULL;
	LLThread 		* TempBiosThread;	

	/* Debug Info */
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]CreateThread",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]CREATE_THREAD : Address %8.8lx",
		Ps2Bios.PS2_Sched.RunningIDThread,MipsCpu.GPR[CPU_REG_A0].lo);
#endif

	// Get the Thread param addr
	ThreadAddr = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;

	// check if I can create a thread
	if ( Ps2Bios.PS2_Sched.TotalThread < BIOS_MAX_THREADS ) {

		// Get the Thread Param structure
		myTh = (  PS2_Thread * )	&MipsMemory [  ThreadAddr & 0x01FFFFFF ];

		// Debug Info
#ifdef THREAD_LOG
		THREAD_LOG("|status      : %8.8lx\n",  myTh->status);
		THREAD_LOG("|entry_add   : %8.8lx\n",  myTh->entry_add);
		THREAD_LOG("|stack_add   : %8.8lx\n",  myTh->stack_add);

		THREAD_LOG("|stackSize   : %8.8lx\n",  myTh->stackSize);
   		THREAD_LOG("|gpReg_add   : %8.8lx\n",  myTh->gpReg_add);       
    	THREAD_LOG("|initPriority: %8.8lx\n",  myTh->initPriority);           
   		THREAD_LOG("|currPriority: %8.8lx\n",  myTh->currentPriority);		           
    	THREAD_LOG("|attr        : %8.8lx\n",  myTh->attr);	
    	THREAD_LOG("|option      : %8.8lx\n",  myTh->option);	
    	THREAD_LOG("|waitType    : %8.8lx\n",  myTh->waitType);
    	THREAD_LOG("|waitId      : %8.8lx\n",  myTh->waitId);		       
    	THREAD_LOG("|wakeupCount : %8.8lx\n",  myTh->wakeupCount);	             
#endif	
	
		// Create the Thread
		TempBiosThread = THREADS_Create( myTh,	THS_DORMANT );
		
		// Set the Thread Status
		TempBiosThread->thread->status = THS_DORMANT;
		TempBiosThread->thread->currentPriority = TempBiosThread->thread->initPriority; 

		// Return the Thread ID
		MipsCpu.GPR[CPU_REG_V0].lo = TempBiosThread->ID;
		
		// Debug Info			
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]CREATE_THREAD : %d OK",Ps2Bios.PS2_Sched.RunningIDThread,TempBiosThread->ID);
#endif

	} else {
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]CREATE_THREAD : FAILED ",Ps2Bios.PS2_Sched.RunningIDThread);
#endif

		// Can't create a thread
		MipsCpu.GPR[CPU_REG_V0].lo = -1;
	}


	MipsCpu.PC = MipsCpu.PC + 4; 
}






void SYSCALL_DeleteThread( ) 					{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]DeleteThread : NI",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]DELETE_THREAD : NI",Ps2Bios.PS2_Sched.RunningIDThread);
#endif

	MipsCpu.halt = 1; 
}
 




void SYSCALL_StartThread( ) 					{ 
	_s16		ThreadID;
	_u32		ThreadArg;

	_u32 		gpreg;
	_u32		*gpreg_addr;
	LLThread 		* TempBiosThread;	


	// Get Syscall param
	ThreadID 	= (_u32)MipsCpu.GPR[CPU_REG_A0].lo;
	ThreadArg	= (_u32)MipsCpu.GPR[CPU_REG_A1].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]StartThread",(long)MipsCpu.PC );
#endif

	// Find the Thread
	TempBiosThread = ( LLThread *) LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );


	if ( TempBiosThread != NULL ) {

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]START_THREAD : %d ( arg : %8.8x ) OK",Ps2Bios.PS2_Sched.RunningIDThread,
		ThreadID,ThreadArg );
#endif


		// Copy the Current Values of GPR & FPU to the Thread*/
		memcpy( TempBiosThread->GPR, MipsCpu.GPR, sizeof( _u128) * 34 );
		memcpy( TempBiosThread->FPR1, MipsCpu.FPR1, sizeof( _u32) * 32 );

		// Set the Entry Addr of the Thread
		TempBiosThread->PC = TempBiosThread->thread->entry_add;

		// Set the Stack Adress of the Thread
		TempBiosThread->GPR[CPU_REG_SP].lo = TempBiosThread->thread->stack_add;
		TempBiosThread->GPR[CPU_REG_S8].lo = TempBiosThread->thread->stack_add;

		// Set the Global Pointer
		TempBiosThread->GPR[CPU_REG_GP].lo = TempBiosThread->thread->gpReg_add;

		// Pass Arg add to the Thread
		TempBiosThread->GPR[CPU_REG_A0].lo = MipsCpu.GPR[CPU_REG_A1].lo;

		// Set the Thread State as READY to be executed
		THREADS_MoveToReady ( TempBiosThread->ID );	
		TempBiosThread->thread->status = THS_READY;
	
	
		// Return the Thread ID
		MipsCpu.GPR[CPU_REG_V0].lo = TempBiosThread->ID;

	} else {
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]START_THREAD : %d ( arg : %8.8x ) FAILED\n",Ps2Bios.PS2_Sched.RunningIDThread,
		ThreadID,ThreadArg );
#endif

		MipsCpu.GPR[CPU_REG_V0].lo = (_s32)-1;
	}

	MipsCpu.PC = MipsCpu.PC + 4; 

	// Update the Scheduler
	UpdatePS2Scheduler();
}






void SYSCALL_ExitThread( ) 						{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ExitThread",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]EXIT_THREAD",Ps2Bios.PS2_Sched.RunningIDThread );
#endif

	Ps2Bios.PS2_Sched.Running_Thread->thread->status = THS_DORMANT;

	MipsCpu.PC = MipsCpu.PC + 4; 

	UpdatePS2Scheduler();

} 




void SYSCALL_ExitDeleteThread( ) 				{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ExitDeleteThread",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]EXIT_DELETE_THREAD",	Ps2Bios.PS2_Sched.RunningIDThread);
#endif

	Ps2Bios.PS2_Sched.Running_Thread->thread->status = THS_DESTROY;

	MipsCpu.PC = MipsCpu.PC + 4; 

	UpdatePS2Scheduler();

}




 
void SYSCALL_TerminateThread( ) 				{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]TerminateThread : NI",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]TERMINATE_THREAD : NI",
		Ps2Bios.PS2_Sched.RunningIDThread);
#endif


	MipsCpu.halt = 1; 
} 
void SYSCALL_iTerminateThread( ) 				{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]iTerminateThread : NI",(long)MipsCpu.PC );
#endif
	MipsCpu.halt = 1;
} 



void SYSCALL_DisableDispatchThread( ) 			{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]DisableDispatchThread : NI",(long)MipsCpu.PC );
#endif
	

	MipsCpu.halt = 1; 
} 





void SYSCALL_EnableDispatchThread( ) 			{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]EnableDispatchThread : NI",(long)MipsCpu.PC );
#endif
	

	MipsCpu.halt = 1; 
} 





void SYSCALL_ChangeThreadPriority( ) 			{ 
	LLThread 		* TempBiosThread = NULL;
	_s16	ThreadID;
	_u32	Priority;

	_u8	NeedUpdateSceduler = FALSE;

	ThreadID = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;
	Priority = (_u32)MipsCpu.GPR[CPU_REG_A1].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ChangeThreadPriority",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]CHANGE_THREAD_PRIORITY : Thr n %d , priority : %d ",
		Ps2Bios.PS2_Sched.RunningIDThread,ThreadID,Priority);
#endif


	TempBiosThread = ( LLThread *) LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );

	if ( TempBiosThread != NULL ) {
		MipsCpu.GPR[CPU_REG_V0].lo = (_u32)TempBiosThread->thread->currentPriority;
		TempBiosThread->thread->currentPriority = Priority;

		NeedUpdateSceduler = TRUE;

/*
		if ( Ps2Bios.PS2_Sched.Running_Thread->thread->currentPriority <= TempBiosThread->thread->currentPriority ) {
			
			switch ( TempBiosThread->thread->status ) {
				case THS_RUN:		break;
			

				case THS_READY: 
					LLThread_AddFirst ( &Ps2Bios.PS2_Sched.Ready_Threads,
						LLThread_Remove ( 	&Ps2Bios.PS2_Sched.Ready_Threads, 
											LLThread_Find ( Ps2Bios.PS2_Sched.All_Threads , TempBiosThread->ID ) ,
											LL_NO_FREE
										)
						);					
					NeedUpdateSceduler = TRUE;
				break;

			case THS_WAIT		:
			case THS_SUSPEND	:
			case THS_WAITSUSPEND:
			case THS_DORMANT	:
	
					LLThread_AddFirst ( &Ps2Bios.PS2_Sched.Ready_Threads,
						LLThread_Remove ( 	&Ps2Bios.PS2_Sched.Other_Threads, 
											LLThread_Find ( Ps2Bios.PS2_Sched.All_Threads , TempBiosThread->ID ) ,
											LL_NO_FREE
										)
						);
			
					NeedUpdateSceduler = TRUE;
				break;

			}

		}
*/

/*
		if ( ThreadID != Ps2Bios.PS2_Sched.RunningIDThread ) {
			// Set the new State
		
			if ( TempBiosThread->thread->status != THS_READY ) {

				THREADS_MoveToReady ( TempBiosThread->ID );
				TempBiosThread->thread->status = THS_READY;
				
			}
			NeedUpdateSceduler = TRUE;
		}
*/
	} else {
		MipsCpu.GPR[CPU_REG_V0].lo = -1;
	}

	



	if ( NeedUpdateSceduler == TRUE ) {
		if ( IntC.ExecutingHandler	== EXECUTING_HANDLER) {
			IntC.ExecutingHandler = NEED_UPDATE_SCHEDULER;		 
		} else {
			UpdatePS2Scheduler(); 
		}
	}

	
	MipsCpu.PC = MipsCpu.PC + 4; 

	
} 
void SYSCALL_iChangeThreadPriority( ) 			{
#ifdef THREAD_LOG 
	THREAD_LOG("%8.8lx:[BIOS]iChangeThreadPriority",(long)MipsCpu.PC );
#endif

	SYSCALL_ChangeThreadPriority( ); 
}    





void SYSCALL_RotateThreadReadyQueue( ) 			{ 
	char temp_string[256];
	_u8	Priority;

	_u8	boucle,result;
	
	Priority = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RotateThreadReadyQueue",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]ROTATE_THREAD : %d \n",Ps2Bios.PS2_Sched.RunningIDThread,Priority );
#endif

	
	THREADS_RotateReadyThread( );

	MipsCpu.GPR[CPU_REG_V0].lo = Priority;

	MipsCpu.PC = MipsCpu.PC + 4; 

	UpdatePS2Scheduler();
} 
void SYSCALL_iRotateThreadReadyQueue( )			{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]iRotateThreadReadyQueue : NI",(long)MipsCpu.PC );
#endif
	MipsCpu.halt = 1; 
}     





void SYSCALL_ReleaseWaitThread( ) 				{ 
#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]ReleaseWaitThread : NI",(long)MipsCpu.PC );
#endif


	MipsCpu.halt = 1; 
} 




void SYSCALL_GetThreadId( ) 					{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]GetThreadId",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]GET_THREAD_ID \n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif
	
	//Return the ID of current Thread
	MipsCpu.GPR[CPU_REG_V0].lo = Ps2Bios.PS2_Sched.RunningIDThread;

	MipsCpu.PC = MipsCpu.PC + 4; 
} 





void SYSCALL_ReferThreadStatus( ) 				{ 

	char 		temp_string[256];
	_s16		ThreadID;
	LLThread 		* TempBiosThread;
	PS2_Thread 	* ParamInfo;
	_u32 		InfoParamAdd;	

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ReferThreadStatus",(long)MipsCpu.PC );
#endif

	
	// Get the Thread ID
	ThreadID = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;

	// Get the return Adress
	InfoParamAdd = (_u32)MipsCpu.GPR[CPU_REG_A1].lo;


#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]REFER_THREAD : n %d \n",Ps2Bios.PS2_Sched.RunningIDThread,ThreadID );
#endif




	TempBiosThread = (LLThread 	*) LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );


	if ( ThreadID == 0 ) {
		TempBiosThread->thread->currentPriority = 128;

	}



	// if the return Adress isn't null
	if ( InfoParamAdd != 0 ) {

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]REFER_THREAD return status at addr: %8.8x \n",Ps2Bios.PS2_Sched.RunningIDThread,InfoParamAdd );
#endif

		ParamInfo = (  PS2_Thread * )	&MipsMemory [  InfoParamAdd & 0x01FFFFFF ];
		memcpy ( ParamInfo, TempBiosThread->thread, sizeof(PS2_Thread) );
			

	}



#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]REFER_THREAD Status : %x \n",Ps2Bios.PS2_Sched.RunningIDThread,TempBiosThread->thread->status );
#endif



	// Return the Status	
	MipsCpu.GPR[CPU_REG_V0].lo = TempBiosThread->thread->status;

	MipsCpu.PC = MipsCpu.PC + 4; 

	UpdatePS2Scheduler();

	//MipsCpu.halt = 1; 

} 




void SYSCALL_iReferThreadStatus( ) 				{ 

#ifdef THREAD_LOG
	THREAD_LOG("%8.8lx:[BIOS]iReferThreadStatus",(long)MipsCpu.PC );
#endif
	SYSCALL_ReferThreadStatus( ); 
}     





void SYSCALL_SleepThread( ) 					{ 
	_s16		ThreadID;
	char 		temp_string[256];

	LLThread 		* TempBiosThread;

	_u32		ReturnValue;
	_u8			NeedUpdateSceduler = FALSE;	


	// Get the Thread ID
	ThreadID = Ps2Bios.PS2_Sched.RunningIDThread;

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]SLEEP_THREAD : n %d \n",Ps2Bios.PS2_Sched.RunningIDThread,ThreadID );
#endif
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SleepThread",(long)MipsCpu.PC );
#endif


	TempBiosThread = (LLThread* )LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );


/*
	// if the Thread wakeup coutn is greater than 0
	if ( TempBiosThread->thread->wakeupCount > 0 ) {
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]SLEEP_THREAD : NO \n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif

		// decrement the wakeup Count
		TempBiosThread->thread->wakeupCount = TempBiosThread->thread->wakeupCount - 1;

	} else {
*/
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]SLEEP_THREAD : done \n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif
		// Set the WAIT state
		TempBiosThread->thread->status = THS_WAIT;

		NeedUpdateSceduler = TRUE;
/*
	}
*/

	MipsCpu.GPR[CPU_REG_V0].lo = Ps2Bios.PS2_Sched.RunningIDThread;	

	INCREMENT_PC();	


	if ( NeedUpdateSceduler == TRUE ) {
		if ( IntC.ExecutingHandler	== EXECUTING_HANDLER) {
			IntC.ExecutingHandler = NEED_UPDATE_SCHEDULER;		 
		} else {
			UpdatePS2Scheduler(); 
		}
	}

	
 
} 






void SYSCALL_WakeupThread( ) { 
	_s16		ThreadID;
	char 		temp_string[256];

	LLThread 		* TempBiosThread = NULL;

	_u32		ReturnValue;
	_u8			NeedUpdateSceduler = FALSE;	


	// Get the Thread ID
	ThreadID = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;


#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]WAKEUP_THREAD : n %d ",Ps2Bios.PS2_Sched.RunningIDThread,ThreadID );
#endif
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]WakeupThread",(long)MipsCpu.PC );
#endif

	if ( ThreadID >= 0 )
		TempBiosThread = ( LLThread * )LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );	

	if ( TempBiosThread != NULL ) {
		
		// if the thread is in WAIT state
		if ( TempBiosThread->thread->status != THS_WAIT ) { 
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]WAKEUP_THREAD ( with WakeUp Count Inc. )",Ps2Bios.PS2_Sched.RunningIDThread);
#endif
			// Increment the thread WakeUp count
			TempBiosThread->thread->wakeupCount = TempBiosThread->thread->wakeupCount + 1;
		} 

		



		if ( ThreadID != Ps2Bios.PS2_Sched.RunningIDThread ) {
			// Set the new State
			THREADS_MoveToReady ( TempBiosThread->ID );
			TempBiosThread->thread->status = THS_READY;

			


			MipsCpu.GPR[CPU_REG_V0].lo = ThreadID;
		} else {

			MipsCpu.GPR[CPU_REG_V0].lo = ThreadID;
		}

	
			NeedUpdateSceduler = TRUE;

	
	} else {
#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]WAKEUP_THREAD ( Error Thread Not Found )",Ps2Bios.PS2_Sched.RunningIDThread);
#endif

		MipsCpu.GPR[CPU_REG_V0].lo = -1;

	}

	INCREMENT_PC();	

	
	if ( NeedUpdateSceduler == TRUE ) {
		if ( IntC.ExecutingHandler	== EXECUTING_HANDLER) {
			IntC.ExecutingHandler = NEED_UPDATE_SCHEDULER;		 
		} else {
			UpdatePS2Scheduler(); 
		}
	}

} 


void SYSCALL_iWakeupThread( ) 					{ 
#ifdef THREAD_LOG
	THREAD_LOG("%8.8lx:[BIOS]iWakeupThread",(long)MipsCpu.PC );
#endif
	SYSCALL_WakeupThread();
}     






void SYSCALL_CancelWakeupThread( ) 				{ 

	_s16		ThreadID;
	char 		temp_string[256];

	LLThread 		* TempBiosThread;

	_u32		ReturnValue;
	_u8			NeedUpdateSceduler = FALSE;	

	// Get the Thread ID
	ThreadID = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;


#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]CancelWakeupThread",(long)MipsCpu.PC );
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]CANCEL_WAKEUP_THREAD : n %d \n",Ps2Bios.PS2_Sched.RunningIDThread,ThreadID );
#endif



	TempBiosThread = (LLThread	*)LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );	
		
 
	//  the thread WakeUp count
	MipsCpu.GPR[CPU_REG_V0].lo 			=  TempBiosThread->thread->wakeupCount;
	TempBiosThread->thread->wakeupCount = 0;



	INCREMENT_PC();	



} 





void SYSCALL_SuspendThread( ) 					{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]SuspendThread : NI",(long)MipsCpu.PC );
#endif
	
	MipsCpu.halt = 1; 
} 





void SYSCALL_ResumeThread( ) 					{ 
#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]ResumeThread : NI",(long)MipsCpu.PC );
#endif

	
	MipsCpu.halt = 1; 
} 





void SYSCALL_JoinThread( ) 						{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]JoinThread : NI",(long)MipsCpu.PC );
#endif

	
	MipsCpu.halt = 1; 
} 





void SYSCALL_iResumeThread( ) 					{ 
#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]iResumeThread : NI",(long)MipsCpu.PC );
#endif


	MipsCpu.halt = 1; 
}     




void SYSCALL_iSuspendThread( ) 					{ 

#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]iSuspendThread : NI",(long)MipsCpu.PC );
#endif

	

	MipsCpu.halt = 1; 
}     


void SYSCALL_iCancelWakeupThread( ) 			{ 
#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]iCancelWakeupThread : NI",(long)MipsCpu.PC );
#endif


	
	MipsCpu.halt = 1; 
}     






void SYSCALL_iReleaseWaitThread( ) 				{ 


#ifdef EMU_LOG
	EMU_LOG("%8.8lx:[BIOS]iReleaseWaitThread : NI",(long)MipsCpu.PC );
#endif
	MipsCpu.halt = 1;  
}     










   





