/* 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>


#include "nSX2.h"
#include "PS2Edefs.h"

#include "LL_Threads/LinkedList.h"

/* create a function to avoid this kind of include */ 
#include "timers.h"
 

#include "bios_syscalls.h"

/* LOG Function for the THREAD*/
void THREAD_Log ( char * ftext  , ...   ) {
	char THREAD_LogText[ FILE_NAME_LENGHT ];

	if ( debug_thread == 1 ) {

    	va_list ap;
		va_start (ap, ftext);
  		vsprintf (THREAD_LogText, ftext, ap);
   		va_end (ap);

		sprintf(THREAD_LogText,"%s\n",THREAD_LogText);
		PrintToBIOSOutput 		(  THREAD_LogText ) ;
	}

}



void	THREADS_Init() {

	int 		boucle;				//loop
	int 		result;				//fonction return

	PS2_Thread	* InitPS2Thread;
	LLThread	* TempBiosThread;

	Ps2Bios.PS2_Sched.All_Threads	=	NULL;
	Ps2Bios.PS2_Sched.Running_Thread=	NULL;
	Ps2Bios.PS2_Sched.Ready_Threads	=	NULL;	
	Ps2Bios.PS2_Sched.Other_Threads	=	NULL;

	Ps2Bios.PS2_Sched.MaxID 			= 255;
	Ps2Bios.PS2_Sched.TotalThread 		= 0;
	Ps2Bios.PS2_Sched.RunningIDThread 	= 0;


	InitPS2Thread = ( PS2_Thread * ) malloc ( sizeof( PS2_Thread ) ) ;

	/* Main Thread - Elf Entry */
	InitPS2Thread->status  			= THS_RUN;
	InitPS2Thread->entry_add  		= MipsCpu.EntryPointAdd;
	InitPS2Thread->initPriority		= 1;
	InitPS2Thread->currentPriority	= 1;
	InitPS2Thread->stackSize		= (_u32)0xFFFFF;
	InitPS2Thread->stack_add  		= (_u32)0x81F00000;

	TempBiosThread = THREADS_Create( InitPS2Thread,	THS_RUN );

	Ps2Bios.PS2_Sched.RunningIDThread 	= TempBiosThread->ID;

	//when I call the InitMainThread Syscall , the ThreadID will be 0
	Ps2Bios.PS2_Sched.MaxID 			= 0;	

}

void	THREADS_Shutdown() {

	LLThread_Free( &Ps2Bios.PS2_Sched.All_Threads );


}





/********************************************************************************/
/*	THREAD QUEUE MANAGEMENT														*/
/********************************************************************************/


/* Initialisation of the Queue */
void 	InitQueue ( _s16 * fQueue ) {
	_s16	boucle;
	for ( boucle = 0 ; boucle < BIOS_MAX_THREADS ; boucle++ ) fQueue[boucle] = -1;
	
}

/* Add Thread next the last Thread */
void	AddLastQueue(_s16 * fQueue, _u16 fID) {
	_s16	boucle;
	for ( boucle = 0 ; boucle < BIOS_MAX_THREADS ; boucle++ ) {
		if ( fQueue[boucle] == -1 ) {fQueue[boucle] = fID;break;}
	}
}

/* Add Thread at the begining of the Queue */
void	AddFirstQueue( _s16 * fQueue, _u16 fID ) {
	_s16	boucle;
	for ( boucle = BIOS_MAX_THREADS - 2 ; boucle >= 0 ; boucle-- ) {
		fQueue[boucle + 1] = fQueue[boucle];
	}
	fQueue[0] = fID;
}

/* Remove the Last Thread of the Queue */
void	RemoveLastQueue( _s16 * fQueue ) {
	_s16	boucle;
	for ( boucle = 0 ; boucle < BIOS_MAX_THREADS ; boucle++ ) {
		if ( fQueue[boucle] == -1 ) break;
	}
	fQueue[boucle - 1] = - 1;
}

/* Remove the first Thread of the Queue */
void	RemoveFirstQueue( _s16 * fQueue ) {
	_s16	boucle;
	for ( boucle = 0 ; boucle < BIOS_MAX_THREADS ; boucle++ ) {
		fQueue[boucle] = fQueue[boucle + 1] ;
	}
	fQueue[BIOS_MAX_THREADS - 1] = -1;
}

/* Make a rotation of the Queue */
void	RotateLeftQueue( _s16 * fQueue ) {
	_s16	saved_ID;
	
	saved_ID = fQueue[0];
	RemoveFirstQueue( fQueue );
	AddLastQueue(fQueue, saved_ID );
}

/* Remove a Thread by his ID */
void	RemoveByIDQueue( _s16 * fQueue , _s16 fID) {
	_s16	boucle,boucle2;
	for ( boucle = 0 ; boucle < BIOS_MAX_THREADS ; boucle++ ) {
		if ( fQueue[boucle] == fID ) break;
	}
	for ( boucle2 = boucle ; boucle2 < ( BIOS_MAX_THREADS - 1 ) ; boucle2++ ) {
		fQueue[boucle2] = fQueue[boucle2 + 1] ;
	}
	fQueue[BIOS_MAX_THREADS - 1] = - 1;
}

/* Test if a Thread is present in a Queue */
_u8	IsPresentInQueue( _s16 * fQueue , _s16 fID) {
	_s16	boucle,boucle2;
	_s16	returnValue = FALSE;

	for ( boucle = 0 ; boucle < BIOS_MAX_THREADS ; boucle++ ) {
		if ( fQueue[boucle] == fID ) {
			returnValue = TRUE;
			break;
		}
	}
	return returnValue;
}


/* Debug Function */
void printQueue ( _s16 * fQueue ) {
	char temp_string[256];
	_u8	boucle;

	memset(temp_string,0,256);
	for ( boucle = 0 ; boucle < 10 ; boucle++ ) {
		sprintf(temp_string,"%s %d ",temp_string, fQueue[boucle]) ;
	}
//	PrintToEmuOutput (  temp_string ); 
}



/********************************************************************************/
/*	THREAD SCHEDULER    														*/
/********************************************************************************/


 
void	UpdatePS2Scheduler ( void ) {
	char temp_string[256];	
	_u8	boucle,result;

	_s32	HighPriority;
	_s32	HighPriorityID;

	_s16	WakeUpThread;

	LLThread 		* CurrentThread;
	LLThread 		* ReadyThread;
	
	LLThread 		* TempThread;


	/*  Each Time an Action is done to the Sheduler  , I reset the Thread Timer 	
		else I update the Scheduler with a Timer ( see timer_main.c )
	*/
	Timer[ NSX2_THREAD_TIMER ].Reg->count = 0;
	CLEAR_MODE_OVFF_TIMER( NSX2_THREAD_TIMER );



		/* At least more than 1 Thread must be in use */
		if ( Ps2Bios.PS2_Sched.TotalThread > 1 ) {

#ifdef THREAD_LOG
	THREAD_LOG("------------------------------------------------");
	THREAD_LOG(LLThread_PrintList("Running :", Ps2Bios.PS2_Sched.Running_Thread ));
	THREAD_LOG(LLThread_PrintList("Ready   :", Ps2Bios.PS2_Sched.Ready_Threads ));
	THREAD_LOG(LLThread_PrintList("Others  :", Ps2Bios.PS2_Sched.Other_Threads ));

#endif



			/* if there isn't any Thread running */
			if ( Ps2Bios.PS2_Sched.RunningIDThread == -1 ) {
#ifdef EMU_LOG
				EMU_LOG("(%d)[THREAD]!!! FATAL ERROR - No Thread Running\n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif
				MipsCpu.halt = 1;	

			} else {
			/* A Thread Is running */
				/*...and I update its status */
				switch( Ps2Bios.PS2_Sched.Running_Thread->thread->status ) {
					/* ERROR */
					case THS_READY :
#ifdef EMU_LOG
						EMU_LOG("(%d)[THREAD]!!! FATAL ERROR - The Running Thread  is READY\n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif
						MipsCpu.halt = 1;	
					break;

					/*The thread need to be destroyed */
					case THS_DESTROY :
						if ( Ps2Bios.PS2_Sched.Ready_Threads == NULL ) {
#ifdef EMU_LOG
							EMU_LOG("(%d)[THREAD]!!! FATAL ERROR - The Running Thr Must be Destroyed & No more Ready Thr\n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif
							MipsCpu.halt = 1;
						} else {

							THREADS_DeleteThreads ( Ps2Bios.PS2_Sched.Running_Thread->ID );
		
							ReadyThread = Ps2Bios.PS2_Sched.Ready_Threads;

							/* Restore Thread GPR */
							memcpy( MipsCpu.GPR , ReadyThread->GPR , sizeof( _u128) * 34 );
							memcpy( MipsCpu.FPR1 , ReadyThread->FPR1 , sizeof( _u32) * 32);

							/* Restore the Stack */
							memcpy ( 	&MipsMemory [ ReadyThread->thread->stack_add & 0x01FFFFFF ],
										ReadyThread->Stack,
										ReadyThread->thread->stackSize );		
			
							/* and the PC */
							MipsCpu.PC = ReadyThread->PC;

							ReadyThread->thread->status = THS_RUN ;

							LLThread_RemoveFirst( &Ps2Bios.PS2_Sched.Ready_Threads, LL_NO_FREE );

							Ps2Bios.PS2_Sched.Running_Thread = ReadyThread;
							Ps2Bios.PS2_Sched.RunningIDThread = ReadyThread->ID;

						}

					break;

					
					/* The Running Thread can't run anymore */
					case THS_WAIT			:
					case THS_SUSPEND		:
					case THS_WAITSUSPEND 	:
					case THS_DORMANT		:

						if ( Ps2Bios.PS2_Sched.Ready_Threads == NULL ) {

#ifdef EMU_LOG
							EMU_LOG("(%d)[THREAD]!!! FATAL ERROR - The Running Thr Must Sleep & No more Ready Thr\n",Ps2Bios.PS2_Sched.RunningIDThread );
#endif

						} else {





				/* Get the Thread Param of the Running Thread */
				CurrentThread = Ps2Bios.PS2_Sched.Running_Thread;


				/* Save the GPR State of the Thread */
				memcpy( CurrentThread->GPR, MipsCpu.GPR, sizeof( _u128) * 34 );
				memcpy( CurrentThread->FPR1, MipsCpu.FPR1, sizeof( _u32) * 32 );
	
				/* Save the Stack */
				memcpy ( 	CurrentThread->Stack,
							&MipsMemory [ CurrentThread->thread->stack_add & 0x01FFFFFF ],
							CurrentThread->thread->stackSize );		

				/* and the PC */
				CurrentThread->PC = MipsCpu.PC;

				/* */
				THREADS_MoveToNotReady ( Ps2Bios.PS2_Sched.RunningIDThread );


				THREADS_PutHighPriorityThreadFirst();


				ReadyThread = Ps2Bios.PS2_Sched.Ready_Threads;

				/* Restore Thread GPR */
				memcpy( MipsCpu.GPR, ReadyThread->GPR, sizeof( _u128) * 34 );
				memcpy( MipsCpu.FPR1, ReadyThread->FPR1, sizeof( _u32) * 32);

				/* Restore the Stack */
				memcpy ( 	&MipsMemory [ ReadyThread->thread->stack_add & 0x01FFFFFF ],
							ReadyThread->Stack,
							ReadyThread->thread->stackSize );		
			
				/* and the PC */
				MipsCpu.PC = ReadyThread->PC;

				ReadyThread->thread->status = THS_RUN ;

				LLThread_RemoveFirst( &Ps2Bios.PS2_Sched.Ready_Threads, LL_NO_FREE );

				Ps2Bios.PS2_Sched.Running_Thread = ReadyThread;
				Ps2Bios.PS2_Sched.RunningIDThread = ReadyThread->ID;


						}

			
					break;

					/* Run */
					default:

						if ( Ps2Bios.PS2_Sched.Ready_Threads != NULL ) {
					
							
				// Get the Thread Param of the Running Thread 
				CurrentThread = Ps2Bios.PS2_Sched.Running_Thread;
		

				// Save the GPR State of the Thread 
				memcpy( CurrentThread->GPR, MipsCpu.GPR, sizeof( _u128) * 34 );
				memcpy( CurrentThread->FPR1, MipsCpu.FPR1, sizeof( _u32) * 32 );

				// Save the Stack 
				memcpy ( 	CurrentThread->Stack,
							&MipsMemory [ CurrentThread->thread->stack_add & 0x01FFFFFF ],
							CurrentThread->thread->stackSize );		

	
				// and the PC 
				CurrentThread->PC = MipsCpu.PC;

				CurrentThread->thread->status = THS_READY;

				
				LLThread_AddEnd		( &Ps2Bios.PS2_Sched.Ready_Threads, CurrentThread );


				THREADS_PutHighPriorityThreadFirst();


				ReadyThread = Ps2Bios.PS2_Sched.Ready_Threads;

				// Restore Thread GPR 
				memcpy( MipsCpu.GPR , ReadyThread->GPR , sizeof( _u128) * 34 );
				memcpy( MipsCpu.FPR1 , ReadyThread->FPR1 , sizeof( _u32) * 32);

				// Restore the Stack 
				memcpy ( 	&MipsMemory [ ReadyThread->thread->stack_add & 0x01FFFFFF ],
							ReadyThread->Stack,
							ReadyThread->thread->stackSize );		

			
				// and the PC 
				MipsCpu.PC = ReadyThread->PC;

				ReadyThread->thread->status = THS_RUN ;

				Ps2Bios.PS2_Sched.Running_Thread = ReadyThread;
				Ps2Bios.PS2_Sched.RunningIDThread = ReadyThread->ID;


				LLThread_Remove ( &Ps2Bios.PS2_Sched.Ready_Threads, ReadyThread , LL_NO_FREE);

						}

		

					break;

				}
				


			}
#ifdef THREAD_LOG

	THREAD_LOG(LLThread_PrintList("Running :", Ps2Bios.PS2_Sched.Running_Thread ));
	THREAD_LOG(LLThread_PrintList("Ready   :", Ps2Bios.PS2_Sched.Ready_Threads ));
	THREAD_LOG(LLThread_PrintList("Others  :", Ps2Bios.PS2_Sched.Other_Threads ));
	THREAD_LOG("------------------------------------------------");
#endif



		} // end of more than 0 Thread




}





/*************************************************************************************/

/*
Create A new Tread
Param is the PS2 Thread and must be created before calling this function
return the BIOS Thread pointer
*/
LLThread*	THREADS_Create( PS2_Thread * fPS2_Thread,	_u32 fState ) {

	LLThread	*TempThread;

	/* Create the Thread */
	TempThread = LLThread_CreateItem( fPS2_Thread );

	if ( TempThread == NULL ) {

		return NULL;
	}

	TempThread->Stack = (char*) malloc ( fPS2_Thread->stackSize );


	LLThread_AddEnd_ALL( &Ps2Bios.PS2_Sched.All_Threads, TempThread );


	switch ( fState ) {

		case THS_RUN : Ps2Bios.PS2_Sched.Running_Thread = TempThread;						break;
	
		case THS_READY:	LLThread_AddFirst( &Ps2Bios.PS2_Sched.Ready_Threads, TempThread );	break;
		
		default : LLThread_AddFirst( &Ps2Bios.PS2_Sched.Other_Threads, TempThread );		break;


	}	

	TempThread->ID = Ps2Bios.PS2_Sched.MaxID;

	Ps2Bios.PS2_Sched.MaxID++;
	Ps2Bios.PS2_Sched.TotalThread++;

	return TempThread;

}

void	THREADS_DeleteThreads ( _u32 fID ) {

	LLThread	*TempThread;

	TempThread = LLThread_Find 	( Ps2Bios.PS2_Sched.All_Threads , fID );

	if ( TempThread != NULL ) {

		if ( TempThread == Ps2Bios.PS2_Sched.Running_Thread ) {
			THREADS_MoveToNotReady ( TempThread->ID );
		}

		LLThread_Remove ( &Ps2Bios.PS2_Sched.Ready_Threads, TempThread ,LL_NO_FREE);	
		LLThread_Remove ( &Ps2Bios.PS2_Sched.Other_Threads, TempThread ,LL_NO_FREE);
		LLThread_Remove ( &Ps2Bios.PS2_Sched.All_Threads, TempThread ,LL_WITH_FREE);

		Ps2Bios.PS2_Sched.TotalThread--;
		
	}

}



void	THREADS_MoveToNotReady ( _u32	fID ) {

	LLThread	*TempThread = NULL;

	TempThread = LLThread_Find 	( Ps2Bios.PS2_Sched.All_Threads , fID );


	if ( TempThread != NULL ) {
		/* if the Thread is the Running Thread */
		if ( TempThread->ID == Ps2Bios.PS2_Sched.Running_Thread->ID ) {
			Ps2Bios.PS2_Sched.Running_Thread = NULL;
			Ps2Bios.PS2_Sched.RunningIDThread = -1;
		} else {
			LLThread_Remove ( &Ps2Bios.PS2_Sched.Ready_Threads, TempThread ,LL_NO_FREE);
		}

		LLThread_AddEnd	( &Ps2Bios.PS2_Sched.Other_Threads, TempThread );

	}


}
 
void	THREADS_MoveToReady ( _u32	fID ) {
	_s32	HighPriority;
	_s32	HighPriorityID;



	LLThread	*TempThread;
	_s32 		status;

	TempThread = LLThread_Find 	( Ps2Bios.PS2_Sched.All_Threads , fID );

	if ( TempThread != NULL ) {

		status = TempThread->thread->status;

		switch ( status ) {

			case THS_RUN		:
				Ps2Bios.PS2_Sched.Running_Thread = NULL;
				Ps2Bios.PS2_Sched.RunningIDThread = -1;

				LLThread_AddEnd ( &Ps2Bios.PS2_Sched.Ready_Threads,
					LLThread_Find ( Ps2Bios.PS2_Sched.All_Threads , fID )
				);
			break;
				

			case THS_READY		:	break;

			case THS_WAIT		:
			case THS_SUSPEND	:
			case THS_WAITSUSPEND:
			case THS_DORMANT	:
	
					LLThread_AddFirst ( &Ps2Bios.PS2_Sched.Ready_Threads,
						LLThread_Remove ( 	&Ps2Bios.PS2_Sched.Other_Threads, 
											LLThread_Find ( Ps2Bios.PS2_Sched.All_Threads , fID ) ,
											LL_NO_FREE
										)
						);


		
			break;
		
			

		}

	}


}


void	THREADS_RotateReadyThread( void ) {

	LLThread_AddEnd	( &Ps2Bios.PS2_Sched.Ready_Threads,
		LLThread_RemoveFirst( &Ps2Bios.PS2_Sched.Ready_Threads, LL_NO_FREE )
	);

}


void	THREADS_PutHighPriorityThreadFirst( void ) {

	_s32	HighPriority;
	_s32	HighPriorityID;

	LLThread 		* TempThread;


	if ( Ps2Bios.PS2_Sched.Ready_Threads == NULL ) return;



	HighPriority = Ps2Bios.PS2_Sched.Ready_Threads->thread->currentPriority;
	HighPriorityID = -1;

	TempThread = Ps2Bios.PS2_Sched.Ready_Threads;

	while ( TempThread != NULL ) {
		if ( TempThread->thread->currentPriority > HighPriority ) {
			HighPriorityID = 	TempThread->ID;
			HighPriority = TempThread->thread->currentPriority;
	
		}

		TempThread = TempThread->next;
	}



	if ( HighPriorityID != -1 ) {

		LLThread_AddFirst ( &Ps2Bios.PS2_Sched.Ready_Threads,
				LLThread_Remove ( 	&Ps2Bios.PS2_Sched.Ready_Threads, 
									LLThread_Find ( Ps2Bios.PS2_Sched.All_Threads , HighPriorityID ) ,
									LL_NO_FREE
								)
			);					

	}

}



