/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <windows.h>
#include <commctrl.h>	

#include <stdlib.h>
#include <stdio.h>
 
#include "resource.h"
#include "GUI_MainWindowCallbacks.h"
#include "Translation.h"
#include "GUI_SettingsSheets.h"
#include "GUI_About.h"
#include "GUI_LogWindow.h"

#include "main.h"
#include "nSX2.h"

/* nSX2 Win Maindow Class Name */
const char      g_szClassName[] = "nSX2Class";

HINSTANCE 	App_hInstance ;

HBITMAP  	_hbitmap_bg;

HWND 		h_nSX2wnd;


char cCurDir[MAX_PATH];


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
    LPSTR lpCmdLine, int nCmdShow)
{
    WNDCLASSEX wc;
    HWND hwnd;
    MSG Msg;
    
    char szWindowTitle[50];
    int statwidths[] = {200, 400,-1};


    App_hInstance = hInstance ;
    
    

    sprintf(szWindowTitle,"nSX2 %1.1d.%2.2d.%1.1d", nSX2_VERSION_MAJ,nSX2_VERSION_MIN,nSX2_VERSION_BUILD);
    
    //Step 1: Registering the Window Class
    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_NSX2_ICON));
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_szClassName;
    wc.hIconSm       = (HICON)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_NSX2_ICON), IMAGE_ICON, 16, 16, 0);


    if(!RegisterClassEx(&wc))
    {
        MessageBox(NULL, "Window Registration Failed!", "Error!",
            MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    // Step 2: Creating the Window
    // Was CreateWindowEx... - Scull
    h_nSX2wnd = CreateWindow(g_szClassName,
                             szWindowTitle,
                             WS_EX_CLIENTEDGE | WS_OVERLAPPED | WS_CLIPCHILDREN |
                             WS_CLIPSIBLINGS | WS_SYSMENU | WS_MINIMIZEBOX,
                             CW_USEDEFAULT, CW_USEDEFAULT, 540, 340,
		             NULL,NULL,hInstance,NULL);



    if(h_nSX2wnd == NULL)
    {
        MessageBox(NULL, "Window Creation Failed!", "Error!",
            MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    ShowWindow(h_nSX2wnd, nCmdShow);
    UpdateWindow(h_nSX2wnd);
    
    // Create Status bar

    hStatus = CreateWindowEx(0, STATUSCLASSNAME, NULL,
				WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP, 0, 0, 0, 0,
				h_nSX2wnd, (HMENU)IDC_MAIN_STATUS, GetModuleHandle(NULL), NULL);

    SendMessage(hStatus, SB_SETPARTS, sizeof(statwidths)/sizeof(int), (LPARAM)statwidths);


    // Step 3: The Message Loop
    while(GetMessage(&Msg, NULL, 0, 0) > 0)
    {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);

		

    }
    return Msg.wParam;
}

extern HANDLE g_event;

/* Main Window Event Handler */
LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    char tempText[MAX_PATH];

    switch(msg)
    {
		/* On the Create of the Main Window */
    	case WM_CREATE:
    	{
        	HMENU hMenu, hSubMenu, hLangMenu;
        	HICON hIcon, hIconSm;


			/* Menu Creation */
        	hMenu = CreateMenu();

			/* FILE MENU */
        	hSubMenu = CreatePopupMenu();
  			AppendMenu(hSubMenu, MF_STRING, ID_FILE_LOAD_ISO, "Run &CDVD");      
			AppendMenu(hSubMenu, MF_STRING, ID_FILE_LOAD_ELF, "Run &Elf");
			AppendMenu(hSubMenu, MF_SEPARATOR, 0, 0);
			hLangMenu = CreatePopupMenu();
                        AppendMenu(hLangMenu, MF_STRING , ID_LANG_ENGLISH, "English");
			AppendMenu(hSubMenu, MF_STRING | MF_POPUP, (UINT)hLangMenu, "&Language");
                        AppendMenu(hSubMenu, MF_SEPARATOR, 0, 0);
 			AppendMenu(hSubMenu, MF_STRING, ID_FILE_EXIT, "&Exit");
        	        AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenu, "&File");

/*
			// BIOS DEBUG ITEM
			AppendMenu(hSubMenu, MF_SEPARATOR, 0, 0);
			AppendMenu(hSubMenu, MF_STRING, ID_FILE_LOAD_BIOS, "Run &Bios");
*/


        	
            SetupLanguages(hLangMenu);
        	

			/* EMULATION MENU */
        	        hSubMenu = CreatePopupMenu();
        	        AppendMenu(hSubMenu, MF_STRING, ID_EMU_START, "&Go");   
        	        AppendMenu(hSubMenu, MF_STRING, ID_EMU_PAUSE, "&Pause"); 
	    	        AppendMenu(hSubMenu, MF_STRING, ID_EMU_RESET, "&Reset");
	    	        AppendMenu(hSubMenu, MF_STRING, ID_EMU_STOP, "&Stop");
        	        AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenu, "&Emulation");

		
			//EnableMenuItem(hSubMenu, ID_EMU_PAUSE,  MF_GRAYED );
			EnableMenuItem(hSubMenu, ID_EMU_RESET,  MF_GRAYED );
			//EnableMenuItem(hSubMenu, ID_EMU_STOP,   MF_GRAYED );


			/* DEBUG MENU */
                       	hSubMenu = CreatePopupMenu();
        	        AppendMenu(hSubMenu, MF_STRING, ID_DEBUG_INIT, "&Init"); 
        	        AppendMenu(hSubMenu, MF_STRING, ID_DEBUG_VIEW, "&View");   
        	        //AppendMenu(hSubMenu, MF_STRING, ID_DEBUG_CLOSE, "&Close"); 
        	        AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenu, "&Debug");
			//EnableMenuItem(hSubMenu, ID_DEBUG,  MF_GRAYED );


			/* PLUGINS MENU */
        	        hSubMenu = CreatePopupMenu();
        	        AppendMenu(hSubMenu, MF_STRING, ID_SETTING_LOG, "&Log"); 
        	        AppendMenu(hSubMenu, MF_STRING, ID_SETTING_CLEAR_LOG, "&Clear Log");   
			AppendMenu(hSubMenu, MF_STRING, ID_SETTING_PLUGIN, "&Plugins");
                        AppendMenu(hSubMenu, MF_SEPARATOR, 0, 0);
                        AppendMenu(hSubMenu, MF_STRING, ID_SETTING_GENERAL, "&Emulator");   
        	        AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenu, "&Settings");
			//EnableMenuItem(hSubMenu, ID_SETTING_PLUGIN,  MF_GRAYED );


			/* HELP MENU */
        	        hSubMenu = CreatePopupMenu();
        	        AppendMenu(hSubMenu, MF_STRING, ID_HELP_ABOUT, "&About");   
        	        AppendMenu(hSubMenu, MF_STRING, ID_HELP_NSX2, "&nSX2");   
        	        AppendMenu(hMenu, MF_STRING | MF_POPUP, (UINT)hSubMenu, "&Help");

			/* Attach Menu to the Main Window */
        	        SetMenu(hwnd, hMenu);

			/* Load Background Bitmap from the ressource */ 
			_hbitmap_bg = LoadBitmap(GetModuleHandle(NULL), MAKEINTRESOURCE(IDB_NSX2_LOGO));


			if (_hbitmap_bg == 0)
      			MessageBox(NULL, "Cannot load bitmap from resources", "Error!",	 MB_OK);

			OpenOutputFile( );

			/* Create the Log Window */
			GUI_CreateLogWindow( hwnd );

 			
			/* Load nSX2 Plugins */
			nSX2_Init();

			/* Debuger Initialisation */
			GUI_OpenDebuger( App_hInstance );
			
			TranslateMenu(GetMenu(hwnd),hwnd);
 
			/* Fix bug with other window getting Focus */
			SetForegroundWindow(hwnd);
			
			
 			HideLogWindow();			
    	}
    	break;
		/* End of Create of the Main Window */


		/* On Paint the Main Window */
		case WM_PAINT:
	    {
			BITMAP bm;
			PAINTSTRUCT ps;

        	        HDC hdc = BeginPaint(hwnd, &ps);

        	        HDC hdcMem = CreateCompatibleDC(hdc);
        	        HBITMAP hbmOld = SelectObject(hdcMem, _hbitmap_bg);

        	        GetObject(_hbitmap_bg, sizeof(bm), &bm);

              	        BitBlt(hdc, 0, 0, bm.bmWidth, bm.bmHeight, hdcMem, 0, 0, SRCCOPY);

        	        SelectObject(hdcMem, hbmOld);
        	        DeleteDC(hdcMem);

        	        EndPaint(hwnd, &ps);
    	}
    	break;


		/* Menu Specific Message */
        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                
                case ID_FILE_LOAD_ELF: 
                    GUI_LoadElf( hwnd );
                break;

                case ID_FILE_LOAD_ISO:
                    GUI_LoadISO( hwnd );
                break;

                case ID_FILE_LOAD_BIOS:
                    GUI_LoadBIOS( hwnd );
                break;

                case ID_SETTING_CLEAR_LOG:
                    ClearLogWindow();
                break;
               
                case ID_FILE_EXIT:
                    PostMessage(hwnd, WM_CLOSE, 0, 0);
                break;

                case ID_EMU_START:
                    GUI_EmulationRUN( hwnd );
                break;

                case ID_EMU_RESET:
                break;

		case ID_EMU_PAUSE:
                    GUI_EmulationPAUSE( hwnd );
                break;
                
                case ID_EMU_STOP:
                    GUI_SendStopMainLoop( TRUE );
                break;

                case ID_DEBUG_INIT:
                    GUI_LoadNothing( hwnd );
                break;

		case ID_DEBUG_VIEW:
                    GUI_ViewDebuger( hwnd );
                break;

                case ID_SETTING_LOG:
                    ShowHideLogWindow();
                break;
	
		case ID_SETTING_PLUGIN:
                    GUI_SettingPlugins( hwnd );
                break; 
				
		case ID_SETTING_GENERAL:
                    ChangeSettings( hwnd );
                break;

		case ID_HELP_ABOUT:
                    OpenAboutDialog( hwnd );
                break;

		case ID_HELP_NSX2:
                    sprintf(tempText,"%s\\Docs\\nSX2 User Manual.chm",cCurDir);
   		    ShellExecute(hwnd, "open", tempText, NULL, NULL, SW_SHOWNORMAL); 
		break; 
				
		default :
                    //Language Support  from ID_LANG_ENGLISH to ID_LANG_ENGLISH+100
                    if (LOWORD(wParam) >= ID_LANG_ENGLISH && LOWORD(wParam) <= (ID_LANG_ENGLISH + 100)) {
		        SelectLang(hwnd,LOWORD(wParam));		     
                        TranslateMenu(GetMenu(hwnd),hwnd);
                    }

            }
        break;

        /* On Resize the Window*/
        case WM_SIZE:
               SendMessage(hStatus, WM_SIZE, 0, 0);  
        break;      
        
       
	/* On Close the Window */
        case WM_CLOSE:
               DestroyWindow(hwnd);
        break;

	/* On Destroy the Window */
        case WM_DESTROY:
            /* NOTA : Crash if the user click "Cross" on the system bar while running */
            StopEmulation();
	    	DeleteObject(_hbitmap_bg);
	    	CloseLogWindow();
            PostQuitMessage(0);
        break;

	/* Pass Event to Default Manager */
        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }
    return 0;
}

int SetStatusText (char *Text)
{
    SendMessage(hStatus, SB_SETTEXT, 0, (LPARAM)Text );
	UpdateWindow(h_nSX2wnd);
}

